"""Generate a multi page documentation for documentation.eccenca.com."""

import re
from pathlib import Path

import click
from click import Command, Group

EDIT_NOTE = "<!-- This file was generated - DO NOT CHANGE IT MANUALLY -->"


def get_icon_for_command_group(full_name: str) -> str:
    """Return in mkdocs material icon term for a command group.

    To look for icons: https://squidfunk.github.io/mkdocs-material/reference/icons-emojis/
    """
    return {
        "admin": "material/key-link",
        "admin acl": "material/server-security",
        "admin metrics": "material/chart-line-variant",
        "admin store": "material/database-outline",
        "admin user": "material/account-cog",
        "admin client": "material/account-cog",
        "admin workspace": "material/folder-multiple-outline",
        "admin workspace python": "material/language-python",
        "config": "material/cog-outline",
        "dataset": "eccenca/artefact-dataset",
        "dataset resource": "eccenca/artefact-file",
        "project": "eccenca/artefact-project",
        "project variable": "material/variable-box",
        "query": "eccenca/application-queries",
        "graph": "eccenca/artefact-dataset-eccencadataplatform",
        "graph validation": "octicons/verified-16",
        "vocabulary": "eccenca/application-vocabularies",
        "vocabulary cache": "eccenca/application-vocabularies",
        "workflow": "eccenca/artefact-workflow",
        "workflow scheduler": "material/calendar",
    }.get(full_name, "octicons/cross-reference-24")


def get_tags_for_command_group(full_name: str) -> str:
    """Get list of tags for a command group name as markdown head section."""
    # pylint: disable=consider-using-join
    tags = {
        "admin": ["cmemc"],
        "admin metrics": ["cmemc"],
        "admin store": ["SPARQL", "cmemc"],
        "admin user": ["Keycloak", "Security", "cmemc"],
        "admin client": ["Keycloak", "Security", "cmemc"],
        "admin acl": ["Security", "cmemc"],
        "admin workspace": ["cmemc"],
        "admin workspace python": ["Python", "cmemc"],
        "config": ["Configuration", "cmemc"],
        "dataset": ["cmemc"],
        "dataset resource": ["cmemc"],
        "project": ["Project", "cmemc"],
        "project variable": ["Variables", "cmemc"],
        "query": ["SPARQL", "cmemc"],
        "graph": ["KnowledgeGraph", "cmemc"],
        "graph validation": ["KnowledgeGraph", "Validation", "cmemc"],
        "vocabulary": ["Vocabulary", "cmemc"],
        "vocabulary cache": ["Vocabulary", "cmemc"],
        "workflow": ["Workflow", "cmemc"],
        "workflow scheduler": ["Automate", "cmemc"],
    }.get(full_name, ["cmemc"])
    markdown = "tags:\n"
    for tag in tags:
        markdown += f"  - {tag}\n"
    return markdown


def get_enhanced_help_text_markdown(
    command_or_group: click.Command | click.Group, skip_first_line: bool = False
) -> str:
    """Add ammunition and code ticks where possible."""
    text = str(command_or_group.help)
    first_line = command_or_group.get_short_help_str(limit=200)
    help_text_ = ""
    for line in text.splitlines():
        if skip_first_line and line.strip().startswith(first_line):
            continue
        help_text_ = f"{help_text_}\n\n" if line == "" else f"{help_text_} {line.rstrip().strip()}"
    help_text = ""
    for _ in help_text_.splitlines():
        line = _
        line = line.strip()
        if line.startswith("Example:"):
            # add EXAMPLE as code
            line = line.replace("Example:", "$")
            line = f"""```shell-session title="Example"
{line}
```
"""
        elif line.startswith("Warning:"):
            # add WARNING admonition
            line = line.replace("Warning: ", "")
            line = f"""!!! warning
    {line}
"""
        elif line.startswith("Note:"):
            # add NOTE admonition
            line = line.replace("Note: ", "")
            line = f"""!!! note
    {line}
"""
        else:
            # plain - non-admonition text will get some auto code ticks

            # surround placeholder, such as {{xxx}} with code ticks (`)
            line = re.sub(r"({{[a-zA-Z\-]+}})", r"`\1`", line)
            # surround short options, such as -p, by code ticks (`)
            line = re.sub(r" (-[a-z])", r"`\1`", line)
            # surround options, such as --raw, by code ticks (`)
            line = re.sub(r"(--[a-z\-]+)", r"`\1`", line)
            # surround parameters, such as GRAPH_URI, by code ticks (`)
            line = re.sub(r"([A-Z]+(_[A-Z]+)+)", r"`\1`", line)
        help_text += f"{line}\n"
    return help_text


def get_commands_for_table_recursive(
    ctx: click.core.Context,
    command_group: click.Command | click.Group,
    table_data: list[dict],
    prefix: str,
    group_name: str,
) -> list[dict]:
    """Get flat list of dicts, representing a command each."""
    new_table_data: list[dict] = []
    commands = command_group.commands  # type: ignore[union-attr]
    for key in command_group.commands:  # type: ignore[union-attr]
        item = commands[key]
        if isinstance(item, Group):
            table_data = get_commands_for_table_recursive(
                ctx,
                item,
                table_data,
                f"{prefix} {group_name}".strip(),
                item.name,  # type: ignore[arg-type]
            )
            continue
        if isinstance(item, Command):
            command_name = item.name
            group_link = f"{prefix}/{group_name}/index.md".replace(" ", "/")
            if group_link.startswith("/"):
                group_link = group_link[1:]
            command_anchor = f"{prefix}-{group_name}".replace(" ", "-")
            command_anchor += f"-{command_name}"
            if command_anchor.startswith("-"):
                command_anchor = command_anchor[1:]
            command_link = f"{group_link}#{command_anchor}"
            new_command = {
                "command_name": command_name,
                "command_link": command_link,
                "group_name": f"{prefix} {group_name}".strip(),
                "group_link": group_link,
                "command_description": item.get_short_help_str(limit=200),
            }
            new_table_data.append(new_command)

    return table_data + new_table_data


def get_markdown_for_index(ctx: click.core.Context, commands: dict) -> str:
    """Create the Markdown text for the command reference index.md."""
    header_md = f"""---
title: "cmemc: Command Reference"
description: "This page lists all commands with its short descriptions."
icon: octicons/cross-reference-24
tags:
  - Reference
  - cmemc
---
# Command Reference
{EDIT_NOTE}

!!! info

    cmemc is organized as a tree of command groups, each with a set of
    commands. You can access the command groups in the table as well as in the
    navigation on the left. You can access the commands directly from the table
    or by visiting a command group page first.
"""
    # get commands
    table_data: list[dict] = []
    for key in commands:
        item = commands[key]
        if isinstance(item, Group):
            table_data = get_commands_for_table_recursive(ctx, item, table_data, "", item.name)  # type: ignore[arg-type]
    table_data = sorted(table_data, key=lambda d: d["group_name"])

    rows_string = ""
    for _ in table_data:
        rows_string += f"| [{_['group_name']}]({_['group_link']}) "
        rows_string += f"| [{_['command_name']}]({_['command_link']}) "
        rows_string += f"| {_['command_description']} |\n"
    commands_md = f"""
| Command Group | Command | Description |
| ------------: | :------ | :---------- |
{rows_string}
"""

    return header_md + commands_md


def create_multi_page_documentation(ctx: click.core.Context, directory: str) -> None:
    """Create a multi page reference manual for documentation.eccenca.com.

    Returns: None
    """
    commands = ctx.command.commands  # type: ignore[attr-defined]
    for key in commands:
        item = commands[key]
        if isinstance(item, Group):
            create_group_manual_dir_recursive(item, ctx, prefix=directory, full_name=key)
    # create the main index.md
    Path(directory).mkdir(parents=True, exist_ok=True)
    file_name = f"{directory}/index.md"
    with Path(file_name).open(mode="w", encoding="UTF-8") as text_file:
        text_file.write(get_markdown_for_index(ctx, commands))


def create_group_manual_dir_recursive(
    command_group: click.Command | click.Group, ctx: click.core.Context, prefix: str, full_name: str
) -> None:
    """Create documentation directory (recursive)."""
    name = command_group.name
    directory = f"{prefix}/{name}"
    create_group_index_md(ctx, command_group, directory, full_name)
    commands = command_group.commands  # type: ignore[union-attr]
    for key in commands:
        item = commands[key]
        if isinstance(item, Group):
            create_group_manual_dir_recursive(
                item, ctx=ctx, prefix=directory, full_name=f"{full_name} {item.name}".strip()
            )


def create_group_index_md(
    ctx: click.core.Context,
    command_group: click.Command | click.Group,
    directory: str,
    full_name: str,
) -> None:
    """Create the index.md of command group."""
    Path(directory).mkdir(parents=True, exist_ok=True)
    file_name = f"{directory}/index.md"

    help_text = get_enhanced_help_text_markdown(command_group)
    # config help has now text but too much lists
    if command_group.name == "config":
        help_text = f"```text\n{command_group.help}\n```"

    with Path(file_name).open(mode="w", encoding="UTF-8") as text_file:
        text_file.write(
            f"""---
title: "cmemc: Command Group - {full_name}"
description: "{command_group.get_short_help_str(limit=200)}"
icon: {get_icon_for_command_group(full_name)}
{get_tags_for_command_group(full_name)}---
# {full_name} Command Group
{EDIT_NOTE}

{help_text}

"""
        )
        for key in command_group.commands:  # type: ignore[union-attr]
            item = command_group.commands[key]  # type: ignore[union-attr]
            if isinstance(item, Group):
                continue
            if isinstance(item, Command):
                text_file.write(get_markdown_for_command(ctx, item, full_name))


def get_markdown_for_command(
    ctx: click.core.Context, command: click.Command, group_name: str
) -> str:
    """Create markdown for a single command."""
    formatter = ctx.make_formatter()
    command.format_help(ctx, formatter)

    # the help text is everything without options and usages
    help_text = get_enhanced_help_text_markdown(command, True)

    # the options text is the documentation of the options
    options_text_ = str(formatter.getvalue())
    # remove usage note (beginning)
    options_text_ = options_text_.replace(
        "Usage: cmemc", f"Usage: cmemc {group_name} {command.name}"
    )
    # remove --help option (end)
    options_text_ = options_text_.replace(
        """Options:
  -h, --help  Show this message and exit.""",
        "",
    )
    options_text_ = options_text_.replace("-h, --help", "")
    options_text_ = options_text_.replace("Show this message and exit.", "")
    options_text_ = options_text_.strip()
    # remove help text (ignore everything until Options)
    options_arrived = False
    options_text = ""
    for line in options_text_.splitlines():
        if options_arrived is True:
            options_text = f"{options_text}\n  {line}"
        if line == "Options:":
            options_arrived = True
    if options_text != "":
        options_text = f"""
??? info "Options"
    ```text
{options_text}
    ```
"""

    # the usage text is a singe line of code
    usage_text = (
        str(command.get_usage(ctx))
        .replace("Usage: cmemc", f"$ cmemc {group_name} {command.name}")
        .strip()
    )
    if options_text == "":
        usage_text = usage_text.replace(" [OPTIONS]", "")
    usage_text = f"""```shell-session title="Usage"
{usage_text}
```
"""

    return f"""## {group_name} {command.name}

{command.get_short_help_str(limit=200)}

{usage_text}

{help_text}

{options_text}
"""



"""Utility functions for CLI interface."""

import json
import os
import re
import sys
import unicodedata
from dataclasses import dataclass
from importlib.metadata import version as cmemc_version
from pathlib import Path

import requests
from bs4 import BeautifulSoup
from cmem.cmempy.dp.admin import get_prometheus_data
from cmem.cmempy.dp.proxy.graph import get_graphs_list
from cmem.cmempy.workspace.projects.project import get_projects
from prometheus_client import Metric
from prometheus_client.parser import text_string_to_metric_families

from cmem_cmemc.constants import NAMESPACES
from cmem_cmemc.context import ApplicationContext


def get_version() -> str:
    """Get the current version or SNAPSHOT."""
    return cmemc_version("cmem-cmemc")


def check_python_version(ctx: ApplicationContext) -> None:
    """Check the runtime python version and warn or error."""
    version = sys.version_info
    major_expected = [3]
    minor_expected = [10, 11, 12]
    if version.major not in major_expected:
        ctx.echo_error(f"Error: cmemc can not be executed with Python {version.major}.")
        sys.exit(1)
    if version.minor not in minor_expected and not ctx.is_completing():
        ctx.echo_warning(
            "Warning: You are running cmemc under a non-tested python "
            f"environment ({version.major}.{version.minor})."
        )


def extract_error_message(error: Exception) -> str:
    """Extract a message from an exception."""
    # exceptions with response is HTTPError
    message = type(error).__name__ + ": " + str(error) + "\n"
    try:
        # try to load Problem Details for HTTP API JSON
        details = json.loads(error.response.text)  # type: ignore[attr-defined]
        message += type(error).__name__ + ": "
        if "title" in details:
            message += details["title"] + ": "
        if "detail" in details:
            message += details["detail"]
    except (AttributeError, ValueError):
        # is not json or any other issue, output plain response text
        pass
    return message


def iri_to_qname(iri: str) -> str:
    """Return a qname for an IRI based on well known namespaces.

    In case of no matching namespace, the full IRI is returned.

    Args:
    ----
        iri: the IRI to transform

    Returns: string

    """
    for prefix, namespace in NAMESPACES.items():
        iri = iri.replace(namespace, prefix + ":")
    return iri


def read_rdf_graph_files(directory_path: str) -> list[tuple[str, str]]:
    """Read all files from directory_path and output as tuples.

    The tuple format is (filepath, graph_name),
    for example ("/tmp/rdf.nt", "http://example.com")
    """
    rdf_graphs = []
    for root, _, files in os.walk(directory_path):
        for _file in files:
            full_file_path = Path(root) / _file
            graph_file_name = _file + ".graph"
            full_graph_file_name_path = Path(root) / graph_file_name
            if not _file.endswith(".graph") and Path(full_graph_file_name_path).exists():
                graph_name = read_file_to_string(str(full_graph_file_name_path)).strip()
                rdf_graphs.append((str(full_file_path.resolve()), graph_name))
    return rdf_graphs


def read_file_to_string(file_path: str) -> str:
    """Read file to string."""
    with Path(file_path).open(mode="rb") as _file:
        return _file.read().decode("utf-8")


def get_graphs(writeable: bool = True, readonly: bool = True) -> list:
    """Retrieve list of accessible graphs from DP endpoint.

    readonly=True|writeable=True outputs all graphs
    readonly=False|writeable=True outputs only writeable graphs
    readonly=True|writeable=False outputs graphs without write access
    (but read access)
    """
    all_graphs = get_graphs_list()
    filtered_graphs = []
    for graph in all_graphs:
        if graph["writeable"] and writeable:
            filtered_graphs.append(graph)
        if not graph["writeable"] and readonly:
            filtered_graphs.append(graph)
    return filtered_graphs


def get_graphs_as_dict(writeable: bool = True, readonly: bool = True) -> dict:
    """Get the graph response as dict with IRI as main key."""
    graph_dict = {}
    for graph in get_graphs(writeable=writeable, readonly=readonly):
        graph_dict[graph["iri"]] = graph
    return graph_dict


def convert_uri_to_filename(value: str, allow_unicode: bool = False) -> str:
    """Convert URI to unix friendly filename.

    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert / to underscore. Convert to lowercase.
    Also strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"\.", "_", value.lower())
    value = re.sub(r"/", "_", value.lower())
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def struct_to_table(
    source: str | bool | float | list | dict, table: list | None = None, prefix: str = ""
) -> list:
    """Prepare flat key/value table from a deep structure.

    This function takes structure and creates a flat table out of it,
    key by key, value by value. For each level deeper it prefixes the
    father keys.

    Example input:  {'k1': '1', 'k2': {'k3': '3', 'k4': '4'}}
    Example output: [['k1', '1'], ['k2:k3', '3'], ['k2:k4', '4']]

    Args:
    ----
        source (any): The structure which is transformed into a flat table.
        table (list): The table where the key/value rows will be appended.
        prefix (str): A prefix which is used to indicate the level.

    Returns:
    -------
        The input table extended with rows from the input source.

    """
    if table is None:
        table = []
    if type(source) in (str, bool, int, float):
        table.append([prefix, source])
        return table
    if isinstance(source, dict):
        if len(prefix) != 0:
            prefix = prefix + "."
        for key in source:
            table = struct_to_table(source[key], table, prefix + key)
        return table
    if isinstance(source, list):
        for value in source:
            table = struct_to_table(value, table, prefix)
        return table
    return table


def split_task_id(task_id: str) -> tuple[str, str]:
    """Validate and split cmemc task ID.

    Args:
    ----
        task_id (str): The task ID in the workspace.

    Raises:
    ------
        ValueError: in case the task ID is not splittable

    """
    try:
        project_part = task_id.split(":")[0]
        task_part = task_id.split(":")[1]
    except IndexError as error:
        raise ValueError(f"{task_id} is not a valid task ID.") from error
    return project_part, task_part


def metric_get_labels(family: Metric, clean: bool = True) -> dict[str, list[str]]:
    """Get the labels of a metric family.

    clean: remove keys with only one dimension
    family: the metric family

    Returns: labels as dict
    """
    labels: dict[str, list[str]] = {}
    # build tree structure
    for sample in family.samples:
        for label in sample.labels:
            value = sample.labels[label]
            if label not in labels:
                labels[label] = []
            if value not in labels[label]:
                labels[label].append(value)
    if clean:
        labels = dict(filter(lambda elem: len(elem[1]) > 1, labels.items()))
    return labels


def metrics_get_dict(job_id: str = "DP") -> dict[str, Metric]:
    """Get metrics data as dict."""
    data = {}
    if job_id == "DP":
        for family in text_string_to_metric_families(get_prometheus_data().text):
            data[family.name] = family
        return data
    raise ValueError(f"job name {job_id} unknown.")


def metrics_get_family(job_id: str, metric_id: str) -> Metric:
    """Get family data.

    This function returns a dictionary of metric families.
    """
    try:
        return metrics_get_dict(job_id=job_id)[metric_id]
    except KeyError as error:
        raise ValueError(f"The job {job_id} does not have a metric {metric_id}.") from error


def check_or_select_project(app: ApplicationContext, project_id: str | None = None) -> str:
    """Check for given project, select the first one if there is only one.

    Args:
    ----
        app (ApplicationContext): the click cli app context.
        project_id (str): The project ID.

    Raises:
    ------
        ValueError: if no projects available.
        ValueError: if more than one project is.

    Returns:
    -------
        Maybe project_id if there was no project_id before.

    """
    if project_id is not None:
        return project_id

    projects = get_projects()
    if len(projects) == 1:
        project_name = str(projects[0]["name"])
        app.echo_warning(
            "Missing project (--project) - since there is only one project, "
            f"this is selected: {project_name}"
        )
        return project_name

    if len(projects) == 0:
        raise ValueError(
            "There are no projects available. "
            "Please create a project with 'cmemc project create'."
        )

    # more than one project
    raise ValueError(
        "There is more than one project available so you need to "
        "specify the project with '--project'."
    )


@dataclass
class PublishedPackage:
    """Represents a published package from pypi.org."""

    name: str
    description: str
    version: str
    published: str


def get_published_packages() -> list[PublishedPackage]:
    """Get a scraped list of plugin packages scraped from pypi.org."""
    url = "https://pypi.org/search/?q=%22cmem-plugin-%22"
    soup = BeautifulSoup(requests.get(url, timeout=5).content, "html.parser")
    packages = []
    snippets = soup.find_all("a", class_="package-snippet")
    for _ in snippets:
        name = _.findChildren(class_="package-snippet__name")[0].getText()
        if name == "cmem-plugin-base":
            continue
        if not name.startswith("cmem-plugin-"):
            continue
        description = _.findChildren(class_="package-snippet__description")[0].getText()
        package_version = _.findChildren(class_="package-snippet__version")[0].getText()
        published = _.findChildren(name="time")[0].attrs["datetime"]
        packages.append(
            PublishedPackage(
                name=name, description=description, version=package_version, published=published
            )
        )
    return packages


def convert_iri_to_qname(iri: str, default_ns: str) -> str:
    """Convert IRI to a QName based on the default namespace.

    If the IRI does not match the default namespace, the method returns the full IRI.
    """
    if iri.startswith(default_ns):
        return iri.replace(default_ns, ":")
    return iri


def convert_qname_to_iri(qname: str, default_ns: str) -> str:
    """Convert a QName to an IRI based on the default namespace.

    If the QName does not match the default namespace, the method returns the full IRI.
    """
    if qname.startswith(":"):
        return default_ns + qname[1:]

    return qname



"""graph commands for cmem command line interface."""

import hashlib
import json
import os
from pathlib import Path
from xml.dom import minidom  # nosec
from xml.etree.ElementTree import (  # nosec
    Element,
    SubElement,
    tostring,
)

import click
from click import Argument
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.authorization import refresh
from cmem.cmempy.dp.proxy import graph as graph_api
from cmem.cmempy.dp.proxy.graph import get_graph_import_tree, get_graph_imports
from cmem.cmempy.dp.proxy.sparql import get as sparql_api
from jinja2 import Template
from six.moves.urllib.parse import quote
from treelib import Tree

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.commands.validation import validation_group
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import (
    convert_uri_to_filename,
    get_graphs,
    get_graphs_as_dict,
    iri_to_qname,
    read_rdf_graph_files,
)

UNKNOWN_GRAPH_ERROR = "The graph {} is not accessible or does not exist."


def tuple_to_list(ctx: ApplicationContext, param: Argument, value: tuple) -> list:  # noqa: ARG001
    """Get a list from a tuple

    Used as callback to have mutable values
    """
    return list(value)


def count_graph(graph_iri: str) -> int:
    """Count triples in a graph and return integer."""
    query = "SELECT (COUNT(*) AS ?triples) " + " FROM <" + graph_iri + "> WHERE { ?s ?p ?o }"  # noqa: S608
    result = json.loads(sparql_api(query, owl_imports_resolution=False))
    count = result["results"]["bindings"][0]["triples"]["value"]
    return int(count)


def _get_graph_to_file(  # noqa: PLR0913
    graph_iri: str,
    file_path: str,
    app: ApplicationContext,
    numbers: tuple[int, int] | None = None,
    overwrite: bool = True,
    mime_type: str = "application/n-triples",
) -> None:
    """Request a single graph to a single file (streamed).

    numbers is a tuple of current and count (for output only).
    """
    if Path(file_path).exists():
        if overwrite is True:
            app.echo_warning(f"Output file {file_path} does exist: will overwrite it.")
        else:
            app.echo_warning(f"Output file {file_path} does exist: will append to it.")
    if numbers is not None:
        running_number, count = numbers
        if running_number is not None and count is not None:
            app.echo_info(
                f"Export graph {running_number}/{count}: " f"{graph_iri} to {file_path} ... ",
                nl=False,
            )
    # create and write the .ttl content file
    if overwrite is True:
        triple_file = click.open_file(file_path, "wb")
    else:
        triple_file = click.open_file(file_path, "ab")
    with graph_api.get_streamed(graph_iri, accept=mime_type) as response:
        response.raise_for_status()
        for chunk in response.iter_content(chunk_size=None):
            if chunk:
                triple_file.write(chunk)
        request_headers = response.request.headers
        request_headers.pop("Authorization")
        app.echo_debug(f"cmemc request headers: {request_headers}")
        app.echo_debug(f"server response headers: {response.headers}")
    if numbers is not None:
        app.echo_success("done")


def _get_export_names(app: ApplicationContext, iris: list[str], template: str) -> dict:
    """Get a dictionary of generated file names based on a template.

    Args:
    ----
        app: the context click application
        iris: list of graph iris
        template (str): the template string to use

    Returns:
    -------
        a dictionary with IRIs as keys and filenames as values

    Raises:
    ------
        ValueError in case the template string produces a naming clash,
            means two IRIs result in the same filename

    """
    template_data = app.get_template_data()
    _names = {}
    for iri in iris:
        template_data.update(
            hash=hashlib.sha256(iri.encode("utf-8")).hexdigest(),
            iriname=convert_uri_to_filename(iri),
        )
        _name_created = Template(template).render(template_data) + ".ttl"
        _names[iri] = _name_created
    if len(_names.values()) != len(set(_names.values())):
        raise ValueError(
            "The given template string produces a naming clash. "
            "Please use a different template to produce unique names."
        )
    return _names


def _create_node_label(iri: str, graphs: dict) -> str:
    """Create a label for a node in the tree."""
    if iri not in graphs:
        return "[missing: " + iri + "]"
    title = graphs[iri]["label"]["title"]
    return f"{title} -- {iri}"


def _add_tree_nodes_recursive(tree: Tree, structure: dict, iri: str, graphs: dict) -> Tree:
    """Add all child nodes of iri from structure to tree.

    Call recursively until no child node can be used as parent anymore.

    Args:
    ----
        tree: the graph where to add the nodes
        structure: the result dict of get_graph_import_tree()
        iri: The IRI of the parent
        graphs: the result of get_graphs()

    Returns:
    -------
        the new treelib.Tree object with the additional nodes

    """
    if not tree.contains(iri):
        tree.create_node(tag=_create_node_label(iri, graphs), identifier=iri)
    if iri not in structure:
        return tree
    for child in structure[iri]:
        tree.create_node(tag=_create_node_label(child, graphs), identifier=child, parent=iri)
    for child in structure[iri]:
        if child in structure:
            tree = _add_tree_nodes_recursive(tree, structure, child, graphs)
    return tree


def _add_ignored_nodes(tree: Tree, structure: dict) -> Tree:
    """Add all child nodes as ignored nodes.

    Args:
    ----
        tree: the graph where to add the nodes
        structure: the result dict of get_graph_import_tree()

    Returns:
    -------
        the new treelib.Tree object with the additional nodes

    """
    if len(structure.keys()) > 0:
        for parent in structure:
            for children in structure[parent]:
                tree.create_node(tag="[ignored: " + children + "]", parent=parent)
    return tree


def _get_graphs_filtered(filter_name: str, filter_value: str) -> list[dict]:
    """Get graphs but filtered according to name and value."""
    # not filtered means all graphs
    graphs: list[dict]
    if filter_name is None:
        return get_graphs()
    # check for correct filter names
    possible_filter_names = ("access", "imported-by")
    if filter_name not in possible_filter_names:
        raise ValueError(
            f"{filter_name} is an unknown filter name. " f"Use one of {possible_filter_names}."
        )
    # filter by access condition
    if filter_name == "access":
        if filter_value == "writeable":
            graphs = get_graphs(writeable=True, readonly=False)
        elif filter_value == "readonly":
            graphs = get_graphs(writeable=False, readonly=True)
        else:
            raise ValueError("Filter access is either 'readonly' or 'writeable'.")
    else:
        # default is all graphs
        graphs = get_graphs()
    # filter by imported-by
    if filter_name == "imported-by":
        if filter_value not in get_graphs_as_dict():
            raise ValueError(UNKNOWN_GRAPH_ERROR.format(filter_value))
        imported_graphs = get_graph_imports(filter_value)
        graphs = [_ for _ in graphs if _["iri"] in imported_graphs]
    return graphs


def _add_imported_graphs(iris: list[str], all_graphs: dict) -> list[str]:
    """Get a list of graph IRIs extended with the imported graphs.

    Args:
    ----
        iris: list of graph IRIs
        all_graphs: output of get_graphs_as_dict (dict of all graphs)

    Returns:
    -------
        list of graph IRIs

    """
    extended_list = iris
    for iri in set(iris):
        for _ in get_graph_imports(iri):
            # check if graph exist
            if _ in all_graphs:
                extended_list.append(_)
    return list(set(extended_list))


def _check_and_extend_exported_graphs(
    iris: list[str], all_flag: bool, imported_flag: bool, all_graphs: dict
) -> list[str]:
    """Get a list of IRIs checked and extended.

    Args:
    ----
        iris: List or tuple of given user IRIs
        all_flag: user wants all graphs
        imported_flag: user wants imported graphs as well
        all_graphs: dict of all graphs (get_graph_as_dict())

    Returns:
    -------
        checked and extended list of IRIs

    Raises:
    ------
         ValueError

    """
    # transform given IRI-tuple to a distinct IRI-list
    iris = list(set(iris))
    if len(iris) == 0 and not all_flag:
        raise ValueError(
            "Either provide at least one graph IRI or use the --all option "
            "in order to work with all graphs."
        )
    for iri in iris:
        if iri not in all_graphs:
            raise ValueError(UNKNOWN_GRAPH_ERROR.format(iri))
    if all_flag:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = [str(_) for _ in all_graphs]
    elif imported_flag:
        # does not need be executed in case of --all
        iris = _add_imported_graphs(iris, all_graphs)
    return iris


def _create_xml_catalog_file(app: ApplicationContext, names: dict, output_dir: str) -> None:
    """Create a Protégé suitable XML catalog file.

    Args:
    ----
        app: the cmemc context object
        names: output of _get_export_names()
        output_dir: path where to create the XML file

    """
    file_name = Path(output_dir) / "catalog-v001.xml"
    catalog = Element("catalog")
    catalog.set("prefer", "public")
    catalog.set("xmlns", "urn:oasis:names:tc:entity:xmlns:xml:catalog")
    for name in names:
        uri = SubElement(catalog, "uri")
        uri.set("id", "Auto-Generated Import Resolution by cmemc")
        uri.set("name", name)
        uri.set("uri", names[name])
    parsed_string = minidom.parseString(  # nosec - since source is trusted  # noqa: S318
        tostring(catalog, "utf-8")
    ).toprettyxml(indent="  ")
    file = click.open_file(str(file_name), "w")
    file.truncate(0)
    file.write(parsed_string)
    app.echo_success(f"XML catalog file created as {file_name}.")


def _prepare_tree_output_id_only(iris: list[str], graphs: dict) -> str:
    """Prepare a sorted, de-duplicated IRI list of graph imports."""
    output_iris = []
    for iri in iris:
        # get response for one requested graph
        api_response = get_graph_import_tree(iri)

        # add all imported IRIs to the IRI list
        # add the requested graph as well
        output_iris.append(iri)
        for top_graph in api_response["tree"]:
            output_iris.append(top_graph)
            for sub_graph in api_response["tree"][top_graph]:
                output_iris.append(sub_graph)  # noqa: PERF402

    # prepare a sorted, de-duplicated IRI list of existing graphs
    # and create a line-by-line output of it
    output_iris = sorted(set(output_iris), key=lambda x: x.lower())
    filtered_iris = [iri for iri in output_iris if iri in graphs]
    return "\n".join(filtered_iris[0:]) + "\n"


@click.command(cls=CmemcCommand, name="tree")
@click.option("-a", "--all", "all_", is_flag=True, help="Show tree of all (readable) graphs.")
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the graph importTree API response.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only graph identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the IRIs into other commands. "
    "The output with this option is a sorted, flat, de-duplicated list "
    "of existing graphs.",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.graph_uris,
    callback=tuple_to_list,
)
@click.pass_obj
def tree_command(
    app: ApplicationContext, all_: bool, raw: bool, id_only: bool, iris: list[str]
) -> None:
    """Show graph tree(s) of the owl:imports hierarchy.

    You can output one or more trees of the import hierarchy.

    Imported graphs which do not exist are shown as `[missing: IRI]`.
    Imported graphs which will result in an import cycle are shown as
    `[ignored: IRI]`.
    Each graph is shown with label and IRI.
    """
    graphs = get_graphs_as_dict()
    if not iris and not all_:
        raise ValueError(
            "Either specify at least one graph IRI or use the "
            "--all option to show the owl:imports tree of all graphs."
        )
    if all_:
        iris = [str(_) for _ in graphs]

    for iri in iris:
        if iri not in graphs:
            raise ValueError(UNKNOWN_GRAPH_ERROR.format(iri))

    iris = sorted(iris, key=lambda x: graphs[x]["label"]["title"].lower())

    if raw:
        for iri in iris:
            # direct output of the response for one requested graph
            app.echo_info_json(get_graph_import_tree(iri))
        return

    if id_only:
        app.echo_result(_prepare_tree_output_id_only(iris, graphs), nl=False)
        return

    # normal execution
    output = ""
    for iri in iris:
        # get response for on requested graph
        api_response = get_graph_import_tree(iri)

        tree = _add_tree_nodes_recursive(Tree(), api_response["tree"], iri, graphs)
        tree = _add_ignored_nodes(tree, api_response["ignored"])

        # strip empty lines from the tree.show output
        output += os.linesep.join(
            [
                line
                for line in tree.show(key=lambda x: x.tag.lower(), stdout=False).splitlines()
                if line.strip()
            ]
        )
        output += "\n"
    # result output
    app.echo_result(output, nl=False)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the graphs list API response.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only graph identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the IRIs into other commands.",
)
@click.option(
    "--filter",
    "filter_",
    type=click.Tuple([click.Choice(["access", "imported-by"]), str]),
    shell_complete=completion.graph_list_filter,
    default=[None] * 2,
    help="Filter graphs based on effective access conditions or import "
    "closure. "
    "First parameter CHOICE can be 'access' or 'imported-by'. "
    "The second parameter can be 'readonly' or 'writeable' in case "
    "of 'access' or any readable graph in case of 'imported-by'.",
)
@click.pass_obj
def list_command(
    app: ApplicationContext, raw: bool, id_only: bool, filter_: tuple[str, str]
) -> None:
    """List accessible graphs."""
    filter_name, filter_value = filter_
    graphs = _get_graphs_filtered(filter_name, filter_value)

    if raw:
        app.echo_info_json(graphs)
        return
    if id_only:
        # output a sorted list of graph IRIs
        for graph_desc in sorted(graphs, key=lambda k: k["iri"].lower()):
            app.echo_result(graph_desc["iri"])
        return
    # output a user table
    table = []
    for _ in graphs:
        if len(_["assignedClasses"]) > 0:
            graph_class = iri_to_qname(sorted(_["assignedClasses"])[0])
        else:
            graph_class = ""
        row = [
            _["iri"],
            graph_class,
            _["label"]["title"],
        ]
        table.append(row)
    app.echo_info_table(table, headers=["Graph IRI", "Type", "Label"], sort_column=2)


# pylint: disable=too-many-arguments,too-many-locals
@click.command(cls=CmemcCommand, name="export")
@click.option("-a", "--all", "all_", is_flag=True, help="Export all readable graphs.")
@click.option(
    "--include-imports",
    is_flag=True,
    help="Export selected graph(s) and all graphs which are imported from "
    "these selected graph(s).",
)
@click.option(
    "--create-catalog",
    is_flag=True,
    help="In addition to the .ttl and .graph files, cmemc will create an "
    "XML catalog file (catalog-v001.xml) which can be used by "
    "applications such as Protégé.",
)
@click.option(
    "--output-dir",
    type=click.Path(writable=True, file_okay=False),
    help="Export to this directory.",
)
@click.option(
    "--output-file",
    type=click.Path(writable=True, allow_dash=True, dir_okay=False),
    default="-",
    show_default=True,
    shell_complete=completion.triple_files,
    help="Export to this file.",
)
@click.option(
    "--filename-template",
    "-t",
    "template",
    default="{{hash}}",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.graph_export_templates,
    help="Template for the export file name(s). "
    "Used together with --output-dir. "
    "Possible placeholders are (Jinja2): "
    "{{hash}} - sha256 hash of the graph IRI, "
    "{{iriname}} - graph IRI converted to filename, "
    "{{connection}} - from the --connection option and "
    "{{date}} - the current date as YYYY-MM-DD. "
    "The file suffix will be appended. "
    "Needed directories will be created.",
)
@click.option(
    "--mime-type",
    default="application/n-triples",
    show_default=True,
    type=click.Choice(["application/n-triples", "text/turtle"]),
    help="Define the requested mime type",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.graph_uris,
    callback=tuple_to_list,
)
@click.pass_obj
def export_command(  # noqa: PLR0913
    app: ApplicationContext,
    all_: bool,
    include_imports: bool,
    create_catalog: bool,
    output_dir: str,
    output_file: str,
    template: str,
    mime_type: str,
    iris: list[str],
) -> None:
    """Export graph(s) as NTriples to stdout (-), file or directory.

    In case of file export, data from all selected graphs will be concatenated
    in one file.
    In case of directory export, .graph and .ttl files will be created
    for each graph.
    """
    iris = _check_and_extend_exported_graphs(iris, all_, include_imports, get_graphs_as_dict())

    count: int = len(iris)
    app.echo_debug("graph count is " + str(count))
    if output_dir:
        # output directory set
        app.echo_debug("output is directory")
        # pre-calculate all filenames with the template,
        # in order to output errors on naming clashes as early as possible
        _names = _get_export_names(app, iris, template)
        # create directory
        if not Path(output_dir).exists():
            app.echo_warning("Output directory does not exist: " + "will create it.")
            Path(output_dir).mkdir(parents=True)
        # one .graph, one .ttl file per named graph
        for current, iri in enumerate(iris, start=1):
            # join with given output directory and normalize full path
            triple_file_name = os.path.normpath(Path(output_dir) / _names[iri])
            graph_file_name = triple_file_name + ".graph"
            # output directory is created lazy
            Path(triple_file_name).parent.mkdir(parents=True, exist_ok=True)
            # create and write the .ttl.graph metadata file
            graph_file = click.open_file(graph_file_name, "w")
            graph_file.write(iri + "\n")
            _get_graph_to_file(
                iri, triple_file_name, app, numbers=(current, count), mime_type=mime_type
            )
        if create_catalog:
            _create_xml_catalog_file(app, _names, output_dir)
        return
    # no output directory set -> file export
    if output_file == "-":
        # in case a file is stdout,
        # all triples from all graphs go in and other output is suppressed
        app.echo_debug("output is stdout")
        for iri in iris:
            _get_graph_to_file(iri, output_file, app, mime_type=mime_type)
    else:
        # in case a file is given, all triples from all graphs go in
        app.echo_debug("output is file")
        for current, iri in enumerate(iris, start=1):
            _get_graph_to_file(
                iri,
                output_file,
                app,
                numbers=(current, count),
                overwrite=False,
                mime_type=mime_type,
            )


@click.command(cls=CmemcCommand, name="import")
@click.option(
    "--replace",
    is_flag=True,
    help="Replace / overwrite the graph(s), instead of just adding the triples to the graph.",
)
@click.option(
    "--skip-existing",
    is_flag=True,
    help="Skip importing a file if the target graph already exists in "
    "the store. Note that the graph list is fetched once at the "
    "beginning of the process, so that you can still add multiple "
    "files to one single graph (if it does not exist).",
)
@click.argument(
    "input_path",
    required=True,
    shell_complete=completion.triple_files,
    type=click.Path(allow_dash=False, readable=True),
)
@click.argument("iri", type=click.STRING, required=False, shell_complete=completion.graph_uris)
@click.pass_obj
def import_command(
    app: ApplicationContext,
    input_path: str,
    replace: bool,
    skip_existing: bool,
    iri: tuple[str, ...],
) -> None:
    """Import graph(s) to the store.

    If input is a file, content will be uploaded to the graph identified with the IRI.

    If input is a directory and NO IRI is given, it scans for file-pairs such as
    `xyz.ttl` and `xyz.ttl.graph`, where `xyz.ttl` is the actual triples file and
    `xyz.ttl.graph` contains the graph IRI in the first line: `https://mygraph.de/xyz/`.

    If input is a directory AND a graph IRI is given, it scans for all `*.ttl` files
    in the directory and imports all content to the graph, ignoring the `*.ttl.graph`
    files.

    If the `--replace` flag is set, the data in the graphs will be overwritten,
    if not, it will be added.

    Note: Directories are scanned on the first level only (not recursively).
    """
    # is an array of tuples like this [('path/to/triple.file', 'graph IRI')]
    if replace and skip_existing:
        raise ValueError(
            "The options --replace and --skip-existing are mutually "
            "exclusive, so please remove one of them."
        )
    graphs: list
    if Path(input_path).is_dir():
        if iri is None:
            # in case a directory is the source (and no IRI is given),
            # the graph/nt file structure is crawled
            graphs = read_rdf_graph_files(input_path)
        else:
            # in case a directory is the source AND IRI is given
            graphs = [(file, iri) for file in Path(input_path).glob("*.ttl")]
    elif Path(input_path).is_file():
        if iri is None:
            raise ValueError(
                "Either specify an input file AND a graph IRI or an input directory ONLY."
            )
        graphs = [(input_path, iri)]
    else:
        raise NotImplementedError(
            "Input from special files (socket, FIFO, device file) is not supported."
        )

    existing_graphs = get_graphs_as_dict()
    processed_graphs: set = set()
    count: int = len(graphs)
    current: int = 1
    for triple_file, graph_iri in graphs:
        app.echo_info(
            f"Import file {current}/{count}: " f"{graph_iri} from {triple_file} ... ", nl=False
        )
        if graph_iri in existing_graphs and skip_existing:
            app.echo_warning("skipped")
            continue
        # prevents re-replacing of graphs in a single run
        _replace = False if graph_iri in processed_graphs else replace
        graph_api.post_streamed(graph_iri, triple_file, replace=_replace)
        app.echo_success("replaced" if _replace else "added")
        # refresh access conditions in case of dropped AC graph
        if graph_iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")
        processed_graphs.add(graph_iri)
        current += 1


@click.command(cls=CmemcCommand, name="delete")
@click.option("-a", "--all", "all_", is_flag=True, help="Delete all writeable graphs.")
@click.option(
    "--include-imports",
    is_flag=True,
    help="Delete selected graph(s) and all writeable graphs which are "
    "imported from these selected graph(s).",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.writable_graph_uris,
    callback=tuple_to_list,
)
@click.pass_obj
def delete_command(
    app: ApplicationContext, all_: bool, include_imports: bool, iris: list[str]
) -> None:
    """Delete graph(s) from the store."""
    graphs = get_graphs_as_dict(writeable=True, readonly=False)
    iris = _check_and_extend_exported_graphs(iris, all_, include_imports, graphs)

    count: int = len(iris)
    for current, iri in enumerate(iris, start=1):
        app.echo_info(f"Drop graph {current}/{count}: {iri} ... ", nl=False)
        graph_api.delete(iri)
        app.echo_success("done")
        # refresh access conditions in case of dropped AC graph
        if iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")


@click.command(cls=CmemcCommand, name="open")
@click.argument("iri", type=click.STRING, shell_complete=completion.graph_uris)
@click.pass_obj
def open_command(app: ApplicationContext, iri: str) -> None:
    """Open / explore a graph in the browser."""
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(cls=CmemcCommand, name="count")
@click.option("-a", "--all", "all_", is_flag=True, help="Count all graphs")
@click.option(
    "-s", "--summarize", is_flag=True, help="Display only a sum of all counted graphs together"
)
@click.argument("iris", nargs=-1, type=click.STRING, shell_complete=completion.graph_uris)
@click.pass_obj
def count_command(
    app: ApplicationContext, all_: bool, summarize: bool, iris: tuple[str, ...]
) -> None:
    """Count triples in graph(s).

    This command lists graphs with their triple count.
    Counts do not include imported graphs.
    """
    if not iris and not all_:
        raise ValueError(
            "Either specify at least one graph IRI " "or use the --all option to count all graphs."
        )
    if all_:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = tuple(iri["iri"] for iri in get_graphs())

    count: int
    overall_sum: int = 0
    for iri in iris:
        count = count_graph(iri)
        overall_sum = overall_sum + count
        if not summarize:
            app.echo_result(f"{count!s} {iri}")
    if summarize:
        app.echo_result(str(overall_sum))


@click.group(cls=CmemcGroup)
def graph() -> CmemcGroup:  # type: ignore[empty-body]
    """List, import, export, delete, count, tree or open graphs.

    Graphs are identified by an IRI.

    Note: The get a list of existing graphs,
    execute the `graph list` command or use tab-completion.
    """


graph.add_command(count_command)
graph.add_command(tree_command)
graph.add_command(list_command)
graph.add_command(export_command)
graph.add_command(delete_command)
graph.add_command(import_command)
graph.add_command(open_command)
graph.add_command(validation_group)



"""query commands for cmem command line interface."""

import json
import sys
from hashlib import sha1
from json import JSONDecodeError, load
from pathlib import Path
from shutil import get_terminal_size
from time import sleep, time
from uuid import uuid4

import click
from click import UsageError
from click.shell_completion import CompletionItem
from cmem.cmempy.queries import QUERY_CATALOG, SparqlQuery, cancel_query, get_query_status
from requests import HTTPError

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.object_list import (
    DirectListPropertyFilter,
    DirectValuePropertyFilter,
    Filter,
    ObjectList,
    compare_int_greater_than,
    compare_regex,
)
from cmem_cmemc.utils import extract_error_message, struct_to_table

QUERY_FILTER_TYPES = sorted(["graph", "status", "slower-than", "type", "regex", "trace-id", "user"])
QUERY_FILTER_HELP_TEXT = (
    "Filter queries based on execution status and time. "
    f"First parameter --filter CHOICE can be one of {QUERY_FILTER_TYPES!s}"
    ". The second parameter is based on CHOICE, e.g. int in case of"
    " slower-than, or a regular expression string."
)


class ReplayStatistics:
    """Capture and calculate statistics of a query replay command run."""

    # pylint: disable=too-many-instance-attributes
    run_id: str
    query_minimum: int | None
    query_maximum: int | None
    loop_count: int
    loop_durations: dict[str, int]
    query_durations: dict[str, list[int]]
    current_loop_key: str
    total_duration: int = 0
    query_count: int = 0
    error_count: int = 0
    query_average: float
    catalog = QUERY_CATALOG
    app: ApplicationContext

    def __init__(self, app: ApplicationContext, label: str | None = None):
        """Initialize instance."""
        self.app = app
        self.label = label
        self.run_id = str(uuid4())
        self.loop_count = 0
        self.query_minimum = None
        self.query_maximum = None
        self.loop_durations = {}
        self.query_durations = {}

    def init_loop(self) -> str:
        """Initialize a new loop and reset the loop counts/values."""
        loop_key = str(uuid4())
        self.current_loop_key = loop_key
        self.loop_durations[loop_key] = 0
        self.query_durations[loop_key] = []
        self.loop_count += 1
        self.app.echo_debug(f"Loop {self.loop_count} started: {loop_key}")
        return loop_key

    def _init_query(self, query_: dict) -> SparqlQuery:
        """Initialize query from dict."""
        try:
            if "iri" in query_:
                iri = query_["iri"]
                catalog_entry = self.catalog.get_query(iri)
                if catalog_entry is None:
                    raise ValueError(f"measure_query - query {iri} is not in catalog.")
                return catalog_entry
            query_string = query_["queryString"]
            return SparqlQuery(text=query_string)
        except KeyError as error:
            raise ValueError("measure_query - given input dict has no queryString key.") from error

    def _update_statistic_on_success(self, duration: int) -> None:
        """Update statistics and counters."""
        self.query_durations[self.current_loop_key].append(duration)
        if self.query_minimum is None or duration < self.query_minimum:
            self.query_minimum = duration
        if self.query_maximum is None or duration > self.query_maximum:
            self.query_maximum = duration
        self.total_duration += duration
        self.loop_durations[self.current_loop_key] += duration
        self.query_average = self.total_duration / self.query_count

    def measure_query_duration(self, query_: dict) -> dict:
        """Execute a query and measure the duration."""
        # create the query object
        executed_query = self._init_query(query_)

        # update and return the query object
        if "id" not in query_:
            # create a UUID4, if needed
            query_["id"] = str(uuid4())
        # always re-hash
        query_["queryStringSha1Hash"] = sha1(  # nosec  # noqa: S324
            executed_query.text.encode("utf-8")
        ).hexdigest()
        if "queryString" not in query_:
            # add queryString for catalog queries
            query_["queryString"] = executed_query.text
        if "iri" in query_:
            # use the full IRI in case short one is given
            query_["iri"] = executed_query.url
        if "iri" in query_ and "label" not in query_:
            # extend with label if possible (and needed)
            query_["label"] = executed_query.label
        if "replays" not in query_:
            # create replays list if needed
            query_["replays"] = []

        # init replay object
        this_replay = {
            "runId": self.run_id,
            "loopId": str(self.current_loop_key),
            "replayId": str(uuid4()),
        }
        if self.label is not None:
            this_replay["runLabel"] = self.label

        # execute and measure the query
        try:
            start = round(time() * 1000)
            executed_query.get_results(replace_placeholder=False)
            end = round(time() * 1000)
            duration = end - start

            this_replay["executionStarted"] = str(start)
            this_replay["executionFinished"] = str(end)
            this_replay["executionTime"] = str(duration)

            # update statistics and counters
            self.query_count += 1
            self._update_statistic_on_success(duration)
        except HTTPError as error:
            self.error_count += 1
            this_replay["executionError"] = extract_error_message(error)

        query_["replays"].append(this_replay)
        self.app.echo_debug(
            f"Query {self.query_count + self.error_count} " f"executed: {this_replay['replayId']}"
        )
        return query_

    def create_output(self) -> dict:
        """Create the structure for the output commands."""
        output = {}
        for key, value in vars(self).items():
            if key not in ("current_loop_key", "app"):
                # ignore some object vars on output
                output[key] = value
        loop_average = 0
        loop_minimum = None
        loop_maximum = None
        for loop_duration in self.loop_durations.values():
            loop_average += loop_duration
            if loop_minimum is None or loop_duration < loop_minimum:
                loop_minimum = loop_duration
            if loop_maximum is None or loop_duration > loop_maximum:
                loop_maximum = loop_duration
        output["loop_minimum"] = loop_minimum
        output["loop_maximum"] = loop_maximum
        output["loop_average"] = loop_average / len(self.loop_durations)
        return output

    def output_table(self) -> None:
        """Output a table of the statistic values."""
        table = struct_to_table(self.create_output())
        self.app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)

    def output_json(self) -> None:
        """Output json of the statistic values."""
        self.app.echo_info_json(self.create_output())


def get_queries(ctx: click.Context) -> list[dict]:  # noqa: ARG001
    """Get queries for object list"""
    _: list[dict] = get_query_status()
    return _


def transform_status_names(ctx: Filter, value: str) -> str:  # noqa: ARG001
    """Transform: status names

    FINISHED_SUCCESS -> finished
    FINISHED_* -> *
    others -> same
    all lowercase
    """
    value = value.replace("FINISHED_", "").lower()
    if value == "success":
        value = "finished"
    return value


query_status_list = ObjectList(
    name="queries",
    get_objects=get_queries,
    filters=[
        DirectValuePropertyFilter(
            name="status",
            description="List only queries which have a certain status.",
            property_key="queryExecutionState",
            transform=transform_status_names,
            fixed_completion=[
                CompletionItem("running", help="List only queries which are currently running."),
                CompletionItem(
                    "finished", help="List only queries which are successfully finished."
                ),
                CompletionItem(
                    "error", help="List only queries which were NOT successfully executed."
                ),
                CompletionItem("cancelled", help="List only queries which were cancelled."),
                CompletionItem("timeout", help="List only queries which ran into a timeout."),
            ],
            fixed_completion_only=True,
        ),
        DirectValuePropertyFilter(
            name="type",
            description="List only queries of a certain query type.",
            property_key="type",
        ),
        DirectValuePropertyFilter(
            name="trace-id",
            description="List only queries which have the specified trace ID.",
            property_key="traceId",
        ),
        DirectValuePropertyFilter(
            name="user",
            description="List only queries executed by the specified account (URL).",
            property_key="user",
        ),
        DirectListPropertyFilter(
            name="graph",
            description="List only queries which affected a certain graph (URL).",
            property_key="affectedGraphs",
        ),
        DirectValuePropertyFilter(
            name="slower-than",
            description="List only queries which are slower than X milliseconds.",
            property_key="executionTime",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("5", help="List only queries which are executed slower than 5ms."),
                CompletionItem(
                    "100", help="List only queries which are executed slower than 100ms."
                ),
                CompletionItem(
                    "1000", help="List only queries which are executed slower than 1000ms."
                ),
                CompletionItem(
                    "5000", help="List only queries which are executed slower than 5000ms."
                ),
            ],
        ),
        DirectValuePropertyFilter(
            name="regex",
            description="List only queries which query text matches a regular expression.",
            property_key="queryString",
            compare=compare_regex,
            fixed_completion=[
                CompletionItem(
                    r"http\:\/\/schema.org",
                    help="List only queries which somehow use the schema.org namespace.",
                ),
                CompletionItem(
                    r"http\:\/\/www.w3.org\/2000\/01\/rdf-schema\#",
                    help="List only queries which somehow use the RDF schema namespace.",
                ),
                CompletionItem(
                    r"\\\?value",
                    help="List only queries which are using the ?value projection variable.",
                ),
                CompletionItem(
                    "^CREATE SILENT GRAPH",
                    help="List only queries which start with CREATE SILENT GRAPH.",
                ),
            ],
        ),
    ],
)


def _output_query_status_details(app: ApplicationContext, status_dict: dict) -> None:
    """Output key/value table as well as query string of a query.

    Args:
    ----
        app: application context
        status_dict: The dict from the query status list.

    """
    table = []
    for key in status_dict:
        if key != "queryString":
            row = [key, str(status_dict[key])]
            table.append(row)
    app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)
    app.echo_info("")
    app.echo_info_sparql(status_dict["queryString"])


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only query identifier and no labels or other metadata. "
    "This is useful for piping the ids into other cmemc commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, id_only: bool) -> None:
    """List available queries from the catalog.

    Outputs a list of query URIs which can be used as reference for
    the query execute command.
    """
    queries = QUERY_CATALOG.get_queries().items()
    if id_only:
        # sort dict by short_url - https://docs.python.org/3/howto/sorting.html
        for _, sparql_query in sorted(queries, key=lambda k: k[1].short_url.lower()):
            app.echo_info(sparql_query.short_url)
    else:
        table = []
        for _, sparql_query in queries:
            row = [
                sparql_query.short_url,
                sparql_query.query_type,
                ",".join(sparql_query.get_placeholder_keys()),
                sparql_query.label,
            ]
            table.append(row)
        app.echo_info_table(
            table, headers=["Query URI", "Type", "Placeholder", "Label"], sort_column=3
        )


@click.command(cls=CmemcCommand, name="execute")
@click.argument(
    "QUERIES", nargs=-1, required=True, shell_complete=completion.remote_queries_and_sparql_files
)
@click.option(
    "--accept",
    default="default",
    show_default=True,
    help="Accept header for the HTTP request(s). Setting this to 'default' "
    "means that cmemc uses an appropriate accept header for terminal "
    "output (text/csv for tables, text/turtle for graphs, * otherwise). "
    "Please refer to the Corporate Memory system manual for a list "
    "of accepted mime types.",
)
@click.option(
    "--no-imports",
    is_flag=True,
    help="Graphs which include other graphs (using owl:imports) will be "
    "queried as merged overall-graph. This flag disables this "
    "default behaviour. The flag has no effect on update queries.",
)
@click.option(
    "--base64",
    is_flag=True,
    help="Enables base64 encoding of the query parameter for the "
    "SPARQL requests (the response is not touched). "
    "This can be useful in case there is an aggressive firewall between "
    "cmemc and Corporate Memory.",
)
@click.option(
    "--parameter",
    "-p",
    type=(str, str),
    shell_complete=completion.placeholder,
    multiple=True,
    help="In case of a parameterized query (placeholders with the '{{key}}' "
    "syntax), this option fills all placeholder with a given value "
    "before the query is executed."
    "Pairs of placeholder/value need to be given as a tuple 'KEY VALUE'. "
    "A key can be used only once.",
)
@click.option(
    "--limit",
    type=int,
    help="Override or set the LIMIT in the executed SELECT query. Note that "
    "this option will never give you more results than the LIMIT given "
    "in the query itself.",
)
@click.option("--offset", type=int, help="Override or set the OFFSET in the executed SELECT query.")
@click.option(
    "--distinct", is_flag=True, help="Override the SELECT query by make the result set DISTINCT."
)
@click.option(
    "--timeout", type=int, help="Set max execution time for query evaluation (in milliseconds)."
)
@click.pass_obj
def execute_command(  # noqa: PLR0913
    app: ApplicationContext,
    queries: tuple[str, ...],
    accept: str,
    no_imports: bool,
    base64: bool,
    parameter: tuple[tuple[str, str], ...],
    limit: int,
    offset: int,
    distinct: bool,
    timeout: int,
) -> None:
    """Execute queries which are loaded from files or the query catalog.

    Queries are identified either by a file path, a URI from the query
    catalog, or a shortened URI (qname, using a default namespace).

    If multiple queries are executed one after the other, the first failing
    query stops the whole execution chain.

    Limitations: All optional parameters (e.g. accept, base64, ...) are
    provided for ALL queries in an execution chain. If you need different
    parameters for each query in a chain, run cmemc multiple times and use
    the logical operators && and || of your shell instead.
    """
    # pylint: disable=too-many-arguments
    placeholder = {}
    for key, value in parameter:
        if key in placeholder:
            raise ValueError(
                "Parameter can be given only once, " f"Value for '{key}' was given twice."
            )
        placeholder[key] = value
    app.echo_debug("Parameter: " + str(placeholder))
    for file_or_uri in queries:
        app.echo_debug(f"Start of execution: {file_or_uri} with " f"placeholder {placeholder}")
        executed_query = QUERY_CATALOG.get_query(file_or_uri, placeholder=placeholder)
        if executed_query is None:
            raise ValueError(f"{file_or_uri} is neither a (readable) file nor a query URI.")
        app.echo_debug(
            f"Execute ({executed_query.query_type}): "
            f"{executed_query.label} < {executed_query.url}"
        )
        if accept == "default":
            submitted_accept = executed_query.get_default_accept_header()
            app.echo_debug(f"Accept header set to default value: '{submitted_accept}'")
        else:
            submitted_accept = accept

        results = executed_query.get_results(
            accept=submitted_accept,
            owl_imports_resolution=not no_imports,
            base64_encoded=base64,
            placeholder=placeholder,
            distinct=distinct,
            limit=limit,
            offset=offset,
            timeout=timeout,
        )
        app.echo_result(results)


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "QUERIES", nargs=-1, required=True, shell_complete=completion.remote_queries_and_sparql_files
)
@click.pass_obj
def open_command(app: ApplicationContext, queries: tuple[str, ...]) -> None:
    """Open queries in the editor of the query catalog in your browser.

    With this command, you can open (remote) queries from the query catalog in
    the query editor in your browser (e.g. in order to change them).
    You can also load local query files into the query editor, in order to
    import them into the query catalog.

    The command accepts multiple query URIs or files which results in
    opening multiple browser tabs.
    """
    for file_or_uri in queries:
        opened_query = QUERY_CATALOG.get_query(file_or_uri)
        if opened_query is None:
            raise ValueError(f"{file_or_uri} is neither a (readable) file nor a query URI.")
        open_query_uri = opened_query.get_editor_url()
        app.echo_debug(f"Open {file_or_uri}: {open_query_uri}")
        click.launch(open_query_uri)


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only query identifier and no labels or other metadata. "
    "This is useful for piping the ids into other cmemc commands.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON response of the query status API.")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=query_status_list.get_filter_help_text(),
    shell_complete=query_status_list.complete_values,
)
@click.argument("query_id", required=False, type=click.STRING)
@click.pass_context
def status_command(
    ctx: click.Context,
    id_only: bool,
    raw: bool,
    filter_: tuple[tuple[str, str]],
    query_id: str,
) -> None:
    """Get status information of executed and running queries.

    With this command, you can access the latest executed SPARQL queries
    on the DataPlatform. These queries are identified by UUIDs and listed
    ordered by starting timestamp.

    You can filter queries based on status and runtime in order to investigate
    slow queries. In addition to that, you can get the details of a specific
    query by using the ID as a parameter.
    """
    width, height = get_terminal_size((120, 20))
    max_query_string_width = width - 46 - 1
    app: ApplicationContext = ctx.obj
    app.echo_debug(f"Terminal size = {width} x {height}")

    if query_id:
        query_status_list.add_filter(
            DirectValuePropertyFilter(name="id", property_key="id", description="id")
        )
        queries = query_status_list.apply_filters(ctx=ctx, filter_=[("id", query_id)])
        query_status_list.remove_filter("id")
    else:
        queries = query_status_list.apply_filters(ctx=ctx, filter_=filter_)

    if query_id and len(queries) == 0:
        raise UsageError(f"Query with ID '{query_id}' does not exist (anymore).")

    if raw:
        app.echo_info_json(queries)
        return

    if id_only:
        for _ in queries:
            app.echo_info(_["id"])
        return

    if query_id:
        _output_query_status_details(app, queries[0])
        return
    table = []
    for _ in queries:
        query_id = _["id"]
        query_execution_time = str(_.get("executionTime", ""))
        query_string = " ".join(_["queryString"].splitlines())
        if len(query_string) > max_query_string_width:
            query_string = query_string[0:max_query_string_width] + "…"
        row = [query_id, query_execution_time, query_string]
        table.append(row)
    app.echo_info_table(table, headers=["Query ID", "Time", "Query String"])


@click.command(cls=CmemcCommand, name="replay")
@click.argument(
    "REPLAY_FILE",
    required=True,
    shell_complete=completion.replay_files,
    type=click.Path(exists=True, allow_dash=False, readable=True, dir_okay=False),
)
@click.option("--raw", is_flag=True, help="Output the execution statistic as raw JSON.")
@click.option(
    "--loops",
    required=False,
    default=1,
    show_default=True,
    type=int,
    help="Number of loops to run the replay file.",
)
@click.option(
    "--wait",
    required=False,
    default=0,
    show_default=True,
    type=int,
    help="Number of seconds to wait between query executions.",
)
@click.option(
    "--output-file",
    required=False,
    shell_complete=completion.replay_files,
    help="Save the optional output to this file. Input and output of the "
    "command can be the same file. The output is written at the end "
    "of a successful command execution. The output can be stdout "
    "('-') - in this case, the execution statistic output is "
    "oppressed.",
    type=click.Path(exists=False, allow_dash=True, writable=True, dir_okay=False),
)
@click.option("--run-label", type=click.STRING, help="Optional label of this replay run.")
@click.pass_obj
def replay_command(  # noqa: PLR0913
    app: ApplicationContext,
    replay_file: str,
    raw: bool,
    loops: int,
    wait: int,
    output_file: str,
    run_label: str,
) -> None:
    """Re-execute queries from a replay file.

    This command reads a REPLAY_FILE and re-executes the logged queries.
    A REPLAY_FILE is a JSON document which is an array of JSON objects with
    at least a key `queryString` holding the query text OR a key `iri`
    holding the IRI of the query in the query catalog.
    It can be created with the `query status` command.

    Example: query status --raw > replay.json

    The output of this command shows basic query execution statistics.

    The queries are executed one after another in the order given in the
    input REPLAY_FILE. Query placeholders / parameters are ignored. If a
    query results in an error, the duration is not counted.

    The optional output file is the same JSON document which is used as input,
    but each query object is annotated with an additional `replays` object,
    which is an array of JSON objects which hold values for the
    replay|loop|run IDs, start and end time as well as duration and
    other data.
    """
    if loops <= 0:
        raise UsageError("Please set a positive loops integer value (>=1).")
    try:
        with Path(replay_file).open(encoding="utf8") as _:
            input_queries = load(_)
    except JSONDecodeError as error:
        raise ValueError(f"File {replay_file} is not a valid JSON document.") from error
    if len(input_queries) == 0:
        raise ValueError(f"File {replay_file} contains no queries.")
    app.echo_debug(f"File {replay_file} contains {len(input_queries)} queries.")

    statistic = ReplayStatistics(app=app, label=run_label)
    for _loop in range(loops):
        statistic.init_loop()
        for _ in input_queries:
            statistic.measure_query_duration(_)
            if wait > 0:
                sleep(wait)

    if output_file:
        if output_file == "-":
            app.echo_info_json(input_queries)
            return
        with Path(output_file).open(mode="w", encoding="utf-8") as output:
            json.dump(input_queries, output, ensure_ascii=False, indent=2)

    if raw:
        statistic.output_json()
    else:
        statistic.output_table()


@click.command(cls=CmemcCommand, name="cancel")
@click.argument("query_id", required=True, type=click.STRING)
@click.pass_context
def cancel_command(ctx: click.Context, query_id: str) -> None:
    """Cancel a running query.

    With this command, you can cancel a running query.
    Depending on the backend triple store, this will result in a
    broken result stream (stardog, neptune and virtuoso) or a valid
    result stream with incomplete results (graphdb)
    """
    app = ctx.obj
    app.echo_info(f"Cancel query {query_id} ... ", nl=False)
    query_status_list.add_filter(
        DirectValuePropertyFilter(name="id", property_key="id", description="id")
    )
    queries = query_status_list.apply_filters(
        ctx=ctx, filter_=[("id", query_id), ("status", "running")]
    )
    if not queries:
        app.echo_error("not known or not running (anymore)")
        sys.exit(1)
    cancel_query(queries[0]["id"])
    app.echo_success("done")


@click.group(cls=CmemcGroup)
def query() -> CmemcGroup:  # type: ignore[empty-body]
    """List, execute, get status or open SPARQL queries.

    Queries are identified either by a file path, a URI from the query
    catalog or a shortened URI (qname, using a default namespace).

    One or more queries can be executed one after the other with the
    execute command. With open command you can jump to the query editor in your
    browser.

    Queries can use a mustache like syntax to specify placeholder for
    parameter values (e.g. {{resourceUri}}). These parameter values need to
    be given as well, before the query can be executed (use the -p option).

    Note: In order to get a list of queries from the query catalog, execute
    the `query list` command or use tab-completion.
    """


query.add_command(execute_command)
query.add_command(list_command)
query.add_command(open_command)
query.add_command(status_command)
query.add_command(replay_command)
query.add_command(cancel_command)



"""Generate a single page documentation Markdown."""

import click


def print_manual(ctx: click.core.Context) -> None:
    """Output the complete manual.

    Returns: None
    """
    ctx.obj.echo_info(
        """# Command Reference

This section lists the help texts of all commands as a reference
and to search for it.
"""
    )
    print_group_manual_recursive(ctx.command, ctx=ctx)


def print_group_manual_recursive(
    command_group: click.Command | click.Group, ctx: click.core.Context, prefix: str = ""
) -> None:
    """Output the help text of a command group (recursive)."""
    commands = command_group.commands  # type: ignore[union-attr]
    for key in commands:
        command = commands[key]
        formatter = ctx.make_formatter()
        if isinstance(command, click.Group):
            ctx.obj.echo_info(f"## Command group: {prefix}{key}\n")
            ctx.obj.echo_info("```")
            command.format_help(ctx, formatter)
            ctx.obj.echo_info(formatter.getvalue().rstrip("\n"))
            ctx.obj.echo_info("```\n")

            print_group_manual_recursive(command, ctx=ctx, prefix=f"{prefix}{key} ")
        elif isinstance(command, click.Command):
            ctx.obj.echo_info(f"### Command: {prefix}{key}\n")
            ctx.obj.echo_info("```text")
            command.format_help(ctx, formatter)
            ctx.obj.echo_info(formatter.getvalue().rstrip("\n"))
            ctx.obj.echo_info("```\n")
        else:
            pass



"""cmemc constants"""

NAMESPACES = {
    "void": "http://rdfs.org/ns/void#",
    "di": "https://vocab.eccenca.com/di/",
    "shui": "https://vocab.eccenca.com/shui/",
    "dsm": "https://vocab.eccenca.com/dsm/",
    "owl": "http://www.w3.org/2002/07/owl#",
    "sd": "http://www.w3.org/ns/sparql-service-description#",
}

NS_ACL = "http://eccenca.com/ac/"
NS_USER = "http://eccenca.com/"
NS_GROUP = "http://eccenca.com/"



"""DataIntegration python management commands."""

import sys
from dataclasses import asdict
from re import match

import click
from click import UsageError
from cmem.cmempy.workspace.python import (
    install_package_by_file,
    install_package_by_name,
    list_packages,
    list_plugins,
    uninstall_package,
    update_plugins,
)

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import get_published_packages


def _get_package_id(module_name: str) -> str:
    """Return package identifier."""
    return module_name.split(".")[0].replace("_", "-")


def _looks_like_a_package(package: str) -> bool:
    """Check if a string looks like a package requirement string."""
    if match(r"^[a-zA-Z0-9]+(-[a-zA-Z0-9]+)*((==|<=|>=|>|<).*)?$", package):
        return True
    return False


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "PACKAGE",
    shell_complete=completion.installable_packages,
    type=click.Path(readable=True, allow_dash=False, dir_okay=False),
)
@click.pass_obj
def install_command(app: ApplicationContext, package: str) -> None:
    """Install a python package to the workspace.

    This command is essentially a `pip install` in the remote python
    environment.

    You can install a package by uploading a source distribution
    .tar.gz file, by uploading a build distribution .whl file, or by
    specifying a package name, i.e., a pip requirement specifier with a
    package name available on pypi.org (e.g. `requests==2.27.1`).

    Note: The tab-completion of this command lists only public packages from
    pypi.org and not from additional or changed python package repositories you
    may have configured on the server.
    """
    app.echo_info(f"Install package {package} ... ", nl=False)
    try:
        response = install_package_by_file(package_file=package)
    except FileNotFoundError as error:
        if not _looks_like_a_package(package):
            raise ValueError(
                f"{package} does not look like a package name or requirement "
                "string, and a file with this name also does not exists."
            ) from error
        response = install_package_by_name(package_name=package)

    # DI >= 24.1 has a combine 'output' key, before 24.1 'standardOutput' and 'errorOutput' existed
    output: list[str] = []
    output.extend(response.get("output", "").splitlines())
    output.extend(response.get("standardOutput", "").splitlines())
    output.extend(response.get("errorOutput", "").splitlines())
    for output_line in output:
        app.echo_debug(output_line)

    if response["success"]:
        app.echo_success("done")
    else:
        app.echo_error("error")
        for output_line in output:
            app.echo_error(output_line)
    app.echo_debug("Updated Plugins: " + str(update_plugins()))


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument("PACKAGE_NAME", nargs=-1, shell_complete=completion.installed_package_names)
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="This option removes all installed packages from the system,"
    " leaving only the pre-installed mandatory packages in the environment.",
)
@click.pass_obj
def uninstall_command(app: ApplicationContext, package_name: str, all_: bool) -> None:
    """Uninstall a python packages from the workspace.

    This command is essentially a `pip uninstall` in the remote
    python environment.
    """
    if all_:
        app.echo_info("Wiping Python environment ... ", nl=False)
        uninstall_package(package_name="--all")
        app.echo_success("done")
        app.echo_debug("Updated Plugins: " + str(update_plugins()))
        return

    if not package_name:
        raise UsageError(
            "Either give at least one package name or wipe the whole python"
            " environment with the '--all' option."
        )
    packages = list_packages()
    app.echo_debug(packages)
    for _ in package_name:
        app.echo_info(f"Uninstall package {_} ... ", nl=False)
        if _ not in [package["name"] for package in packages]:
            app.echo_error("not installed")
            app.echo_debug("Updated Plugins: " + str(update_plugins()))
            sys.exit(1)
        response = uninstall_package(package_name=_)
        output: list[str] = []
        output.extend(response.get("output", "").splitlines())
        output.extend(response.get("standardOutput", "").splitlines())
        output.extend(response.get("errorOutput", "").splitlines())
        for output_line in output:
            app.echo_debug(output_line)
        if response["success"]:
            app.echo_success("done")
        else:
            app.echo_error("error")
            app.echo_debug(response.content.decode())
    app.echo_debug("Updated Plugins: " + str(update_plugins()))


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only package identifier. " "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--available",
    is_flag=True,
    help="Instead of listing installed packages, this option lists installable packages"
    " from pypi.org, which are prefixed with 'cmem-plugin-' and so are most likely"
    " Corporate Memory plugin packages.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool, available: bool) -> None:
    """List installed python packages.

    This command is essentially a `pip list` in the remote python environment.

    It outputs a table of python package identifiers with version information.
    """
    if available:
        published_packages = get_published_packages()

        if raw:
            app.echo_info_json([asdict(_) for _ in published_packages])
            return

        if id_only:
            for _ in published_packages:
                app.echo_info(_.name)
            return

        table_published = [
            (_.name, _.version, str(_.published)[:10], _.description) for _ in published_packages
        ]
        app.echo_info_table(
            table_published, headers=["Name", "Version", "Published", "Description"], sort_column=0
        )
        return

    installed_packages = list_packages()
    if raw:
        app.echo_info_json(installed_packages)
        return

    if id_only:
        for package in installed_packages:
            app.echo_info(package["name"])
        return

    table_installed = [(_["name"], _["version"]) for _ in installed_packages]
    app.echo_info_table(table_installed, headers=["Name", "Version"], sort_column=0)


@click.command(cls=CmemcCommand, name="list-plugins")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option("--id-only", is_flag=True, help="Lists only plugin identifier.")
@click.option("--package-id-only", is_flag=True, help="Lists only plugin package identifier.")
@click.pass_obj
def list_plugins_command(
    app: ApplicationContext, raw: bool, id_only: bool, package_id_only: bool
) -> None:
    """List installed workspace plugins.

    This commands lists all discovered plugins.

    Note: The plugin discovery is restricted to package prefix (`cmem-`).
    """
    raw_output = list_plugins()
    try:
        plugins = raw_output["plugins"]  # DI >= 22.1.1 output
    except TypeError:
        plugins = raw_output  # DI <= 22.1 output

    if raw:
        app.echo_info_json(plugins)
        return

    if id_only:
        for plugin in sorted(plugins, key=lambda k: k["id"].lower()):
            app.echo_info(plugin["id"])
        return

    if package_id_only:
        package_ids = set()
        for plugin in plugins:
            package_ids.add(_get_package_id(plugin["moduleName"]))
        for package_id in sorted(package_ids):
            app.echo_info(package_id)
        return

    table = [
        (
            _["id"],
            _get_package_id(_["moduleName"]),
            _["pluginType"],
            _["label"],
        )
        for _ in sorted(plugins, key=lambda k: k["id"].lower())
    ]
    app.echo_info_table(table, headers=["ID", "Package ID", "Type", "Label"], sort_column=0)
    if "error" in raw_output:
        app.echo_error(raw_output["error"])


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "PACKAGE",
    shell_complete=completion.published_package_names,
)
@click.pass_obj
def open_command(app: ApplicationContext, package: str) -> None:
    """Open a package pypi.org page in the browser.

    With this command, you can open the pypi.org page of a published package in your
    browser. From there, you can follow links, review the version history as well
    as the origin of the package, and read the provided documentation.
    """
    full_url = f"https://pypi.org/project/{package}/"
    app.echo_debug(f"Open {package}: {full_url}")
    click.launch(full_url)


@click.group(cls=CmemcGroup)
def python() -> CmemcGroup:  # type: ignore[empty-body]
    """List, install, or uninstall python packages.

    Python packages are used to extend the DataIntegration workspace
    with python plugins. To get a list of installed packages, execute the
    list command.

    Warning: Installing packages from unknown sources is not recommended.
    Plugins are not verified for malicious code.
    """


python.add_command(install_command)
python.add_command(uninstall_command)
python.add_command(list_command)
python.add_command(list_plugins_command)
python.add_command(open_command)



"""DataPlatform store commands for the cmem command line interface."""

import os
from pathlib import Path

import click
from click import UsageError
from cmem.cmempy.dp.admin import create_showcase_data, delete_bootstrap_data, import_bootstrap_data
from cmem.cmempy.dp.admin.backup import get_zip, post_zip

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.context import ApplicationContext


@click.command(cls=CmemcCommand, name="bootstrap")
@click.option(
    "--import",
    "import_",
    is_flag=True,
    help="Delete existing bootstrap data if present and import bootstrap "
    "data which was delivered with Corporate Memory.",
)
@click.option("--remove", is_flag=True, help="Delete existing bootstrap data if present.")
@click.pass_obj
def bootstrap_command(app: ApplicationContext, import_: bool, remove: bool) -> None:
    """Update/Import or remove bootstrap data.

    Use `--import` to import the bootstrap data needed for managing shapes and
    configuration objects. This will remove the old data first.

    Use `--remove` to delete bootstrap data.

    Note: The removal of existing bootstrap data will search for resources which are
    flagged with the isSystemResource property.
    """
    if import_ and remove or not import_ and not remove:
        raise UsageError("Either use the --import or the --remove option.")
    if import_:
        app.echo_info("Update or import bootstrap data ... ", nl=False)
        import_bootstrap_data()
        app.echo_success("done")
        return
    if remove:
        app.echo_info("Remove bootstrap data ... ", nl=False)
        delete_bootstrap_data()
        app.echo_success("done")
        return


@click.command(cls=CmemcCommand, name="showcase")
@click.option(
    "--scale",
    type=click.INT,
    default="10",
    show_default=True,
    help="The scale factor provides a way to set the target size of the "
    "scenario. A value of 10 results in around 40k triples, a value of "
    "50 in around 350k triples.",
)
@click.option(
    "--create",
    is_flag=True,
    help="Delete old showcase data if present and create new showcase data"
    "based on the given scale factor.",
)
@click.option("--delete", is_flag=True, help="Delete existing showcase data if present.")
@click.pass_obj
def showcase_command(app: ApplicationContext, scale: int, create: bool, delete: bool) -> None:
    """Create showcase data.

    This command creates a showcase scenario of multiple graphs including
    integration graphs, shapes, statement annotations, etc.

    Note: There is currently no deletion mechanism for the showcase data, and
    you need to remove the showcase graphs manually (or just remove all
    graphs).
    """
    if not delete and not create:
        raise ValueError("Either use the --create or the --delete flag.")
    if delete:
        raise NotImplementedError(
            "This feature is not implemented yet. " "Please delete the graphs manually."
        )
    if create:
        app.echo_info(f"Create showcase data with scale factor {scale} ... ", nl=False)
        create_showcase_data(scale_factor=scale)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="export")
@click.argument(
    "BACKUP_FILE",
    shell_complete=completion.graph_backup_files,
    required=True,
    type=click.Path(writable=True, allow_dash=False, dir_okay=False),
)
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrite existing files. " "This is a dangerous option, so use it with care.",
)
@click.pass_obj
def export_command(app: ApplicationContext, backup_file: str, overwrite: bool) -> None:
    """Backup all knowledge graphs to a ZIP archive.

    The backup file is a ZIP archive containing all knowledge graphs (one
    Turtle file + configuration file per graph).

    This command will create lots of load on the server.
    It can take a long time to complete.
    """
    if Path(backup_file).exists() and overwrite is not True:
        raise ValueError(
            f"Export file {backup_file} already exists and --overwrite " "option is not used."
        )
    with get_zip() as request:
        request.raise_for_status()
        with Path(backup_file).open(mode="wb") as _:
            requested_size = 1024 * 1024
            byte_counter = 0
            overall_byte_counter = 0
            app.echo_info(f"Exporting graphs backup to {backup_file} ...", nl=False)
            for chunk in request.iter_content(chunk_size=requested_size):
                chunk_size = len(chunk)
                app.echo_debug(f"Got new chuck of {chunk_size} bytes.")
                byte_counter += chunk_size
                overall_byte_counter += chunk_size
                _.write(chunk)
                _.flush()
                os.fsync(_.fileno())
                if byte_counter > requested_size:
                    app.echo_info(".", nl=False)
                    byte_counter = 0
            app.echo_debug(f"Wrote {overall_byte_counter} bytes to {backup_file}.")
            app.echo_success(" done")


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "BACKUP_FILE",
    shell_complete=completion.graph_backup_files,
    required=True,
    type=click.Path(readable=True, exists=True, allow_dash=False, dir_okay=False),
)
@click.pass_obj
def import_command(app: ApplicationContext, backup_file: str) -> None:
    """Restore graphs from a ZIP archive.

    The backup file is a ZIP archive containing all knowledge graphs  (one
    Turtle file + configuration file per graph).

    The command will load a single backup ZIP archive into the triple store
    by replacing all graphs with the content of the Turtle files in the
    archive and deleting all graphs which are not in the archive.

    This command will create lots of load on the server.
    It can take a long time to complete.
    The backup file will be transferred to the server, then unzipped and
    imported graph by graph. After the initial transfer the network
    connection is not used anymore and may be closed by proxies.
    This does not mean that the import failed.
    """
    app.echo_info(f"Importing graphs backup from {backup_file} ...", nl=False)
    request = post_zip(backup_file)
    request.raise_for_status()
    app.echo_success(" done")


@click.group(cls=CmemcGroup)
def store() -> CmemcGroup:  # type: ignore[empty-body]
    """Import, export and bootstrap the knowledge graph store.

    This command group consist of commands to administrate the
    knowledge graph store as a whole.
    """


store.add_command(showcase_command)
store.add_command(bootstrap_command)
store.add_command(export_command)
store.add_command(import_command)



"""metrics commands for cmem command line interface."""

import click
from prometheus_client import Metric

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import (
    metric_get_labels,
    metrics_get_dict,
    metrics_get_family,
    struct_to_table,
)


def _filter_samples(family: Metric, label_filter: tuple[tuple[str, str], ...]) -> list:
    """Filter samples by labels."""
    all_samples = family.samples
    if not label_filter:
        return all_samples
    labels = metric_get_labels(family)
    samples = []
    for sample in all_samples:
        matching_labels = 0
        for name, value in label_filter:
            if name not in labels:
                raise ValueError(
                    f"The metric '{family.name}' does " f"not have a label named '{name}'."
                )
            if value not in labels[name]:
                raise ValueError(
                    f"The metric '{family.name}' does "
                    f"not have a label '{name}' with the value '{value}'."
                )
            if name in sample.labels and sample.labels[name] == value:
                matching_labels += 1
        if matching_labels == len(label_filter):
            # all filter match
            samples.append(sample)
    return samples


# pylint: disable-msg=too-many-arguments
@click.command(cls=CmemcCommand, name="get")
@click.argument("metric_id", required=True, type=click.STRING, shell_complete=completion.metric_ids)
@click.option(
    "--job",
    "job_id",
    type=click.Choice(["DP"]),
    default="DP",
    show_default=True,
    help="The job from which the metrics data is fetched.",
)
@click.option(
    "--filter",
    "label_filter",
    type=(str, str),
    shell_complete=completion.metric_label_filter,
    multiple=True,
    help="A set of label name/value pairs in order to filter the samples "
    "of the requested metric family. Each metric has a different set "
    "of labels with different values. "
    "In order to get a list of possible label names and values, use "
    "the command without this option. The label names are then shown "
    "as column headers and label values as cell values of this column.",
)
@click.option(
    "--enforce-table",
    is_flag=True,
    help="A single sample value will be returned as plain text instead "
    "of the normal table. This allows for more easy integration "
    "with scripts. This flag enforces the use of tabular output, "
    "even for single row tables.",
)
@click.option("--raw", is_flag=True, help="Outputs raw prometheus sample classes.")
@click.pass_obj
def get_command(  # noqa: PLR0913
    app: ApplicationContext,
    metric_id: str,
    job_id: str,
    label_filter: tuple[tuple[str, str], ...],
    raw: bool,
    enforce_table: bool,
) -> None:
    """Get sample data of a metric.

    A metric of a specific job is identified by a metric ID. Possible
    metric IDs of a job can be retrieved with the `metrics list`
    command. A metric can contain multiple samples.
    These samples are distinguished by labels (name and value).
    """
    app.echo_debug(f"Get metrics family '{metric_id}' for job '{job_id}'")
    family = metrics_get_family(job_id, metric_id)
    app.echo_debug(f"Filter family '{metric_id}' with filter '{label_filter}'")
    samples = _filter_samples(family, label_filter)
    if len(samples) == 0:
        raise ValueError(
            "No data - the given label combination filtered out "
            f"all available samples of the metric {metric_id}."
        )

    if raw:
        for sample in samples:
            app.echo_info(str(sample))
        return

    if len(samples) == 1 and enforce_table is not True:
        app.echo_info(str(samples[0].value))
        return

    if family.type == "histogram":
        raise NotImplementedError(
            "Visualization of histogram metrics is not supported yet.\n"
            "Please use the 'metrics inspect' command to get the raw data."
        )

    label_dict = metric_get_labels(family, clean=True)
    table = []
    for sample in samples:
        row = [sample.labels[key] for key in sorted(label_dict.keys())]
        table.append(row)
        row.append(str(sample.value))
    headers = sorted(label_dict.keys())
    headers.append("value")
    app.echo_info_table(table, headers=headers, sort_column=0)


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("metric_id", required=True, type=click.STRING, shell_complete=completion.metric_ids)
@click.option(
    "--job",
    "job_id",
    type=click.Choice(["DP"]),
    default="DP",
    show_default=True,
    help="The job from which the metrics data is fetched.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the table data.")
@click.pass_obj
def inspect_command(app: ApplicationContext, metric_id: str, job_id: str, raw: bool) -> None:
    """Inspect a metric.

    This command outputs the data of a metric.
    The first table includes basic metadata about the metric.
    The second table includes sample labels and values.
    """
    family = metrics_get_family(job_id, metric_id)
    if raw:
        app.echo_info_json({"family": family.__dict__, "labels": metric_get_labels(family)})
        return
    app.echo_info_table(struct_to_table(family.__dict__), headers=["Key", "Value"], sort_column=0)
    app.echo_info_table(
        struct_to_table(metric_get_labels(family)), headers=["Label", "Value"], sort_column=0
    )


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--job",
    "job_id",
    type=click.Choice(["DP"]),
    default="DP",
    show_default=True,
    help="The job from which the metrics data is fetched.",
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists metric identifier only. " "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--raw", is_flag=True, help="Outputs (sorted) JSON dict, parsed from the metrics API output."
)
@click.pass_obj
def list_command(app: ApplicationContext, job_id: str, id_only: bool, raw: bool) -> None:
    """List metrics for a specific job.

    For each metric, the output table shows the metric ID,
    the type of the metric, a count of how many labels (label names)
    are describing the samples (L) and a count of how many samples are
    currently available for a metric (S).
    """
    data = metrics_get_dict(job_id=job_id)
    if raw:
        output = {}
        for key, family in sorted(data.items()):
            output[key] = family.__dict__
        app.echo_info_json(output)
        return
    if id_only:
        for key, _ in sorted(data.items()):
            app.echo_info(key)
        return

    table = []
    for _, family in sorted(data.items()):
        labels = metric_get_labels(family)
        table.append([family.name, family.type, len(labels), len(family.samples)])

    app.echo_info_table(table, headers=["ID", "Type", "L", "S"], sort_column=0)


@click.group(cls=CmemcGroup)
def metrics() -> CmemcGroup:  # type: ignore[empty-body]
    """List and get metrics.

    This command group consists of commands for reading and listing
    internal monitoring metrics of eccenca Corporate Memory. A
    deployment consists of multiple jobs (e.g. DP, DI), which provide
    multiple metric families for an endpoint.

    Each metric family can consist of different samples identified by
    labels with a name and a value (dimensions). A metric has a specific
    type (counter, gauge, summary and histogram) and additional metadata.

    Please have a look at https://prometheus.io/docs/concepts/data_model/
    for further details.
    """


metrics.add_command(get_command)
metrics.add_command(inspect_command)
metrics.add_command(list_command)



"""The main command line interface."""

import contextlib
import os
import sys
import traceback
from importlib.resources import open_text
from subprocess import CalledProcessError  # nosec

import click
import requests.exceptions

from cmem_cmemc import completion
from cmem_cmemc.commands import (
    CmemcGroup,
    admin,
    config,
    dataset,
    graph,
    project,
    query,
    vocabulary,
    workflow,
)
from cmem_cmemc.context import CONTEXT
from cmem_cmemc.exceptions import InvalidConfigurationError
from cmem_cmemc.manual_helper.graph import print_manual_graph
from cmem_cmemc.manual_helper.multi_page import create_multi_page_documentation
from cmem_cmemc.manual_helper.single_page import print_manual
from cmem_cmemc.utils import check_python_version, extract_error_message, get_version

CMEMC_VERSION = get_version()

# this will output a custom zsh completion function
if os.environ.get("_CMEMC_COMPLETE", "") == "zsh_source":
    with open_text("cmem_cmemc", "_cmemc.zsh") as zsh_output:
        CONTEXT.echo_info(zsh_output.read())
    sys.exit(0)

version = sys.version_info
PYTHON_VERSION = f"{version.major}.{version.minor}.{version.micro}"
check_python_version(ctx=CONTEXT)

# set the user-agent environment for the http request headers
os.environ["CMEM_USER_AGENT"] = f"cmemc/{CMEMC_VERSION} (Python {PYTHON_VERSION})"

# https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
CONTEXT_SETTINGS = {"auto_envvar_prefix": "CMEMC", "help_option_names": ["-h", "--help"]}


@click.group(name="cmemc", cls=CmemcGroup, context_settings=CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--connection",
    type=click.STRING,
    shell_complete=completion.connections,
    help="Use a specific connection from the config file.",
)
@click.option(
    "--config-file",
    shell_complete=completion.ini_files,
    type=click.Path(readable=True, allow_dash=False, dir_okay=False),
    default=CONTEXT.config_file_default,
    show_default=True,
    help="Use this config file instead of the default one.",
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress any non-error info messages.")
@click.option(
    "-d", "--debug", is_flag=True, help="Output debug messages and stack traces after errors."
)
@click.version_option(
    version=CMEMC_VERSION,
    message="%(prog)s, version %(version)s, " f"running under python {PYTHON_VERSION}",
)
@click.pass_context
def cli(
    ctx: click.core.Context, debug: bool, quiet: bool, config_file: str, connection: str
) -> None:
    """Eccenca Corporate Memory Control (cmemc).

    cmemc is the eccenca Corporate Memory Command Line Interface (CLI).

    Available commands are grouped by affecting resource type (such as graph,
    project and query).
    Each command and group has a separate --help screen for detailed
    documentation.
    In order to see possible commands in a group, simply
    execute the group command without further parameter (e.g. cmemc project).

    If your terminal supports colors, these coloring rules are applied:
    Groups are colored in white; Commands which change data are colored in
    red; all other commands as well as options are colored in green.

    Please also have a look at the cmemc online documentation:

                        https://eccenca.com/go/cmemc

    cmemc is © 2024 eccenca GmbH, licensed under the Apache License 2.0.
    """
    ctx.obj = CONTEXT
    # hidden feature: 'CMEMC_MANUAL=true cmemc -q config list' will output
    #     the whole markdown manual
    if os.getenv("CMEMC_MANUAL_DIR"):
        create_multi_page_documentation(ctx, str(os.getenv("CMEMC_MANUAL_DIR")))
        ctx.exit()
    # hidden feature: 'CMEMC_MANUAL=true cmemc -q config list' will output
    #     the whole markdown manual
    if os.getenv("CMEMC_MANUAL"):
        print_manual(ctx)
        ctx.exit()
    # hidden feature: 'CMEMC_MANUAL_GRAPH=true cmemc -q config list' will
    # output the documentation graph
    if os.getenv("CMEMC_MANUAL_GRAPH"):
        print_manual_graph(ctx, get_version())
        ctx.exit()
    ctx.obj.set_quiet(quiet)
    ctx.obj.set_debug(debug)
    ctx.obj.set_config_file(config_file)
    try:
        ctx.obj.set_connection(connection)
    except InvalidConfigurationError as error:
        # if config is broken still allow for "config edit"
        # means: do not forward this exception if "config edit"
        if " ".join(sys.argv).find("config edit") == -1:
            raise ValueError("Broken config") from error


cli.add_command(admin.admin)
cli.add_command(config.config)
cli.add_command(dataset.dataset)
cli.add_command(graph.graph)
cli.add_command(project.project)
cli.add_command(query.query)
cli.add_command(vocabulary.vocabulary)
cli.add_command(workflow.workflow)


def main() -> None:
    """Start the command line interface."""
    try:
        cli()  # pylint: disable=no-value-for-parameter
    except (
        OSError,
        CalledProcessError,
        requests.exceptions.HTTPError,
        requests.exceptions.ConnectionError,
        ValueError,
        NotImplementedError,
        KeyError,
    ) as error:
        if CONTEXT.is_completing():
            # if currently autocompleting -> silently die with exit 1
            sys.exit(1)
        CONTEXT.echo_debug(traceback.format_exc())
        CONTEXT.echo_error(extract_error_message(error))
        with contextlib.suppress(
            requests.exceptions.ConnectionError, requests.exceptions.HTTPError
        ):
            CONTEXT.check_versions()
        sys.exit(1)



"""access control"""

import click
import requests.exceptions
from click import Option, UsageError
from cmem.cmempy.dp.authorization.conditions import (
    create_access_condition,
    delete_access_condition,
    fetch_all_acls,
    get_access_condition_by_iri,
    review_graph_rights,
    update_access_condition,
)
from cmem.cmempy.keycloak.user import get_user_by_username, user_groups

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.constants import NS_ACL, NS_GROUP, NS_USER
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import convert_iri_to_qname, convert_qname_to_iri, struct_to_table

# option descriptions
HELP_TEXTS = {
    "name": "A short name or label.",
    "id": "An optional ID (will be an UUID otherwise).",
    "description": "An optional description.",
    "user": "A specific user account required by the access condition.",
    "group": "A membership in a user group required by the access condition",
    "read_graph": "Grants read access to a graph.",
    "write_graph": "Grants write access to a graph (includes read access).",
    "action": "Grants usage permissions to an action / functionality.",
}

PUBLIC_USER_URI = "urn:elds-backend-anonymous-user"
PUBLIC_GROUP_URI = "urn:elds-backend-public-group"

KNOWN_ACCESS_CONDITION_URLS = [PUBLIC_USER_URI, PUBLIC_GROUP_URI]


def _list_to_acl_url(ctx: ApplicationContext, param: Option, value: list) -> list:
    """Option callback which returns a URI for a list of strings.

    or list of URIs, if a tuple comes from click
    or not, if it is already a known URI .... or None
    """
    return [_value_to_acl_url(ctx, param, _) for _ in value]


def _value_to_acl_url(
    ctx: ApplicationContext,  # noqa: ARG001
    param: Option,
    value: str | None,
) -> str | None:
    """Option callback which returns a URI for a string.

    or not, if it is already a known URI .... or None
    """
    if value in KNOWN_ACCESS_CONDITION_URLS:
        return value
    if value == "":
        return ""
    if value is None:
        return None
    match param.name:
        case "groups":
            return f"{NS_GROUP}{value}"
        case "user":
            return f"{NS_USER}{value}"
    return f"{NS_ACL}{value}"


def generate_acl_name(user: str | None, groups: list[str]) -> str:
    """Create an access condition name based on user and group assignments."""
    if len(groups) > 0:
        group_term = "groups" if len(groups) > 1 else "group"
        groups_labels = ", ".join(
            [convert_iri_to_qname(iri=_, default_ns=NS_GROUP)[1:] for _ in groups]
        )
        if user:
            return (
                f"Condition for user {convert_iri_to_qname(iri=user, default_ns=NS_USER)[1:]} "
                f"and {group_term} {groups_labels}"
            )
        return f"Condition for {group_term} {groups_labels}"
    if user:
        return f"Condition for user: {convert_iri_to_qname(iri=user, default_ns=NS_USER)[1:]}"
    return "Condition for ALL users"


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only URIs. This is useful for piping the IDs into other commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool) -> None:
    """List access conditions.

    This command retrieves and lists all access conditions, which are manageable
    by the current account.
    """
    acls = fetch_all_acls()
    if raw:
        app.echo_info_json(acls)
        return
    if id_only:
        for graph in acls:
            app.echo_info(convert_iri_to_qname(iri=graph.get("iri"), default_ns=NS_ACL))
        return
    table = [
        (convert_iri_to_qname(iri=_.get("iri"), default_ns=NS_ACL), _.get("name", "-"))
        for _ in acls
    ]
    app.echo_info_table(table, headers=["URI", "Name"], sort_column=0)


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("access_condition_id", type=click.STRING, shell_complete=completion.acl_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def inspect_command(app: ApplicationContext, access_condition_id: str, raw: bool) -> None:
    """Inspect an access condition.

    Note: access conditions can be listed by using the `acl list` command.
    """
    iri = convert_qname_to_iri(qname=access_condition_id, default_ns=NS_ACL)
    access_condition = get_access_condition_by_iri(iri).json()

    if raw:
        app.echo_info_json(access_condition)
        return

    table = struct_to_table(access_condition)
    app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="create")
@click.option(
    "--name",
    "name",
    type=click.STRING,
    help=HELP_TEXTS["name"],
)
@click.option(
    "--id",
    "id_",
    type=click.STRING,
    help=HELP_TEXTS["id"],
)
@click.option(
    "--description",
    "description",
    type=click.STRING,
    help=HELP_TEXTS["description"],
)
@click.option(
    "--user",
    type=click.STRING,
    shell_complete=completion.acl_users,
    help=HELP_TEXTS["user"],
    callback=_value_to_acl_url,
)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    help=HELP_TEXTS["group"],
    callback=_list_to_acl_url,
)
@click.option(
    "--read-graph",
    "read_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["read_graph"],
)
@click.option(
    "--write-graph",
    "write_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["write_graph"],
)
@click.option(
    "--action",
    "actions",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_actions,
    help=HELP_TEXTS["action"],
)
@click.pass_obj
# pylint: disable-msg=too-many-arguments
def create_command(  # noqa: PLR0913
    app: ApplicationContext,
    name: str,
    id_: str,
    description: str,
    user: str,
    groups: list[str],
    read_graphs: tuple[str],
    write_graphs: tuple[str],
    actions: tuple[str],
) -> None:
    """Create an access condition."""
    if not read_graphs and not write_graphs and not actions:
        raise click.UsageError(
            "Missing access / usage grant. Use at least one of the following options: "
            "--read-graph, --write-graph or --action."
        )
    if not user and not groups:
        app.echo_warning(
            "Access conditions without a user and without a group assignment " "affect ALL users."
        )

    if not name:
        name = generate_acl_name(user=user, groups=groups)

    if not description:
        description = "This access condition was created with cmemc."

    app.echo_info(
        f"Creating access condition '{name}' ... ",
        nl=False,
    )
    create_access_condition(
        name=name,
        static_id=id_,
        description=description,
        user=user,
        groups=groups,
        read_graphs=list(read_graphs),
        write_graphs=list(write_graphs),
        actions=list(actions),
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "access_condition_id",
    nargs=1,
    required=True,
    type=click.STRING,
    shell_complete=completion.acl_ids,
)
@click.option(
    "--name",
    "name",
    type=click.STRING,
    help=HELP_TEXTS["name"],
)
@click.option(
    "--description",
    "description",
    type=click.STRING,
    help=HELP_TEXTS["description"],
)
@click.option(
    "--user",
    type=click.STRING,
    shell_complete=completion.acl_users,
    help=HELP_TEXTS["user"],
    callback=_value_to_acl_url,
)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    help=HELP_TEXTS["group"],
    callback=_list_to_acl_url,
)
@click.option(
    "--read-graph",
    "read_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["read_graph"],
)
@click.option(
    "--write-graph",
    "write_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["write_graph"],
)
@click.option(
    "--action",
    "actions",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_actions,
    help=HELP_TEXTS["action"],
)
@click.pass_obj
# pylint: disable-msg=too-many-arguments
def update_command(  # noqa: PLR0913
    app: ApplicationContext,
    access_condition_id: str,
    name: str,
    description: str,
    user: str,
    groups: list[str],
    read_graphs: tuple[str],
    write_graphs: tuple[str],
    actions: tuple[str],
) -> None:
    """Update an access condition.

    Given an access condition URL, you can change specific options
    to new values.
    """
    iri = convert_qname_to_iri(qname=access_condition_id, default_ns=NS_ACL)
    payload = get_access_condition_by_iri(iri=iri).json()
    app.echo_info(
        f"Updating access condition {payload['name']} ... ",
        nl=False,
    )

    update_access_condition(
        iri=iri,
        name=name,
        description=description,
        user=user,
        groups=groups,
        read_graphs=read_graphs,
        write_graphs=write_graphs,
        actions=actions,
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all access conditions. " "This is a dangerous option, so use it with care.",
)
@click.argument(
    "access_condition_ids",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.acl_ids,
)
@click.pass_obj
def delete_command(app: ApplicationContext, all_: bool, access_condition_ids: list[str]) -> None:
    """Delete access conditions.

    This command deletes existing access conditions from the account.

    Note: Access conditions can be listed by using the `cmemc admin acs list` command.
    """
    if access_condition_ids == () and not all_:
        raise click.UsageError(
            "Either specify at least one access condition ID,"
            " or use the --all option to delete all access conditions."
        )
    if all_:
        access_condition_ids = [_["iri"] for _ in fetch_all_acls()]

    count = len(access_condition_ids)
    for index, _ in enumerate(access_condition_ids, 1):
        app.echo_info(f"Delete access condition {index}/{count}: {_} ... ", nl=False)
        delete_access_condition(iri=convert_qname_to_iri(qname=_, default_ns=NS_ACL))
        app.echo_success("done")


@click.command(name="review")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.argument("user", type=click.STRING, shell_complete=completion.acl_users)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    callback=_list_to_acl_url,
    help="Add groups to the review request (what-if-scenario).",
)
@click.pass_obj
def review_command(app: ApplicationContext, raw: bool, user: str, groups: list[str] | None) -> None:
    """Review grants for a given account.

    This command has two working modes: (1) You can review the access conditions
    of an actual account - this needs access to keycloak and the access condition API,
    (2) You can review the access conditions of an imaginary account with a set of
    freely added groups (what-if-scenario) - this only needs access to the access
    condition API.

    The output of the command is a list of grants the account has based on your input
    and all access conditions loaded in the store. In addition to that, some metadata
    of the account is shown.
    """
    if not groups:
        app.echo_debug("Trying to fetch groups from keycloak.")
        keycloak_user = get_user_by_username(username=user)
        if not keycloak_user:
            raise UsageError(
                "Unknown User or no access to get user info.\n"
                "Use the --group option to assign groups manually (what-if-scenario)."
            )
        try:
            keycloak_user_groups = user_groups(user_id=keycloak_user[0]["id"])
            groups = [f"{NS_GROUP}{_['name']}" for _ in keycloak_user_groups]
        except requests.exceptions.HTTPError as error:
            raise UsageError(
                f"You do not have the permission to retrieve the groups for user {user}"
                " from Keycloak.\n"
                "Use the --group option to assign groups manually (what-if-scenario)."
            ) from error
    app.echo_debug(f"Got groups: {groups}")
    review_info: dict = review_graph_rights(
        account_iri=f"{NS_USER}{user}", group_iris=groups
    ).json()
    review_info["groupIri"] = groups
    if raw:
        app.echo_info_json(review_info)
        return
    table = struct_to_table(review_info)
    app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.group(cls=CmemcGroup)
def acl() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete and modify and review access conditions.

    With this command group, you can manage and inspect access conditions
    in eccenca Corporate Memory. Access conditions are identified by a URL.
    They grant access to knowledge graphs or actions to user or groups.
    """


acl.add_command(list_command)
acl.add_command(inspect_command)
acl.add_command(create_command)
acl.add_command(update_command)
acl.add_command(delete_command)
acl.add_command(review_command)



"""Keycloak client management commands"""

import click
from cmem.cmempy.config import get_keycloak_base_uri, get_keycloak_realm_id
from cmem.cmempy.keycloak.client import (
    generate_client_secret,
    get_client_by_client_id,
    get_client_secret,
    list_open_id_clients,
)

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.context import ApplicationContext

NO_CLIENT_ERROR = (
    "{} is not a valid client account. "
    "Use the 'admin client list' command "
    "to get a list of existing client accounts."
)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only Client ID. " "This is useful for piping the IDs into other commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool) -> None:
    """List client accounts.

    Outputs a list of client accounts, which can be used to get an overview as well
    as a reference for the other commands of the `admin client` command group.

    Note: The list command only outputs clients which have a client secret.
    """
    clients = list_open_id_clients()
    if raw:
        app.echo_info_json(clients)
        return
    if id_only:
        for cnt in clients:
            app.echo_info(cnt["clientId"])
        return
    table = [(_["clientId"], _.get("description", "-")) for _ in clients]
    app.echo_info_table(table, headers=["Client ID", "Description"], sort_column=0)


@click.command(cls=CmemcCommand, name="secret")
@click.argument("client-id", shell_complete=completion.client_ids)
@click.option("--generate", is_flag=True, help="Generate a new secret")
@click.option("--output", is_flag=True, help="Display client secret")
@click.pass_obj
def secret_command(app: ApplicationContext, client_id: str, generate: bool, output: bool) -> None:
    """Get or generate a new secret for a client account.

    This command retrieves or generates a new secret for a client account from a realm.
    """
    if not output and not generate:
        app.echo_info(click.get_current_context().get_help())
        raise ValueError("You need to use '--output' or '--generate' as an option.")

    clients = get_client_by_client_id(client_id)
    if not clients:
        raise ValueError(NO_CLIENT_ERROR.format(client_id))

    if generate:
        if not output:
            app.echo_info(f"Generating a new secret for {client_id} ... ", nl=False)
        credential = generate_client_secret(client_id=clients[0]["id"])
        if not output:
            app.echo_success("done")
    else:
        credential = get_client_secret(client_id=clients[0]["id"])

    if output:
        app.echo_result(credential["value"])


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "client-ids", nargs=-1, required=False, type=click.STRING, shell_complete=completion.client_ids
)
@click.pass_obj
def open_command(app: ApplicationContext, client_ids: tuple[str]) -> None:
    """Open clients in the browser.

    With this command, you can open a client in the keycloak web
    interface in your browser.

    The command accepts multiple client IDs which results in
    opening multiple browser tabs.
    """
    open_client_base_uri = (
        f"{get_keycloak_base_uri()}/admin/master/console/#/" f"{get_keycloak_realm_id()}/clients"
    )
    if not client_ids:
        app.echo_debug(f"Open users list: {open_client_base_uri}")
        click.launch(open_client_base_uri)
    else:
        clients = list_open_id_clients()
        client_id_map = {c["clientId"]: c["id"] for c in clients}
        for _ in client_ids:
            if _ not in client_id_map:
                raise ValueError(NO_CLIENT_ERROR.format(_))
            client_id = client_id_map[_]
            open_user_uri = f"{open_client_base_uri}/{client_id}/settings"

            app.echo_debug(f"Open {_}: {open_user_uri}")
            click.launch(open_user_uri)


@click.group(cls=CmemcGroup)
def client() -> CmemcGroup:  # type: ignore[empty-body]
    """List client accounts, get or generate client account secrets.

    This command group is an opinionated interface to the Keycloak realm of your
    Corporate Memory instance. In order to be able to use the commands in this group,
    the configured cmemc connection account needs to be equipped with the
    `manage-clients` role in the used realm.

    Client accounts are identified by a client ID which is unique in the scope of
    the used realm.

    In case your Corporate Memory deployment does not use the default deployment
    layout, the following additional config variables can be used in your
    connection configuration: `KEYCLOAK_BASE_URI` defaults to
    `{CMEM_BASE_URI}/auth` and locates your Keycloak deployment;
    `KEYCLOAK_REALM_ID` defaults to `cmem` and identifies the used realm.
    """


client.add_command(list_command)
client.add_command(secret_command)
client.add_command(open_command)



"""Declares all cli exceptions."""


class InvalidConfigurationError(ValueError):
    """The configuration given was not found or is broken."""


class ServerError(ValueError):
    """The server reported an error with a process."""



"""graph validation command group"""

import time
from datetime import datetime, timezone

import click
import requests
import timeago
from click import Context, UsageError
from click.shell_completion import CompletionItem
from cmem.cmempy.dp.shacl import validation
from requests import HTTPError
from rich.progress import Progress, SpinnerColumn, TaskID, TimeElapsedColumn

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.completion import _finalize_completion
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.exceptions import ServerError
from cmem_cmemc.object_list import (
    DirectListPropertyFilter,
    DirectValuePropertyFilter,
    ObjectList,
    compare_int_greater_than,
    transform_lower,
)
from cmem_cmemc.utils import struct_to_table


def get_sorted_validations_list(ctx: Context) -> list[dict]:  # noqa: ARG001
    """Get a sorted list of validation objects (aggregations)"""
    objects = validation.get_all_aggregations()
    return sorted(objects, key=lambda o: str(o.get("executionStarted", "SCHEDULED")))


validations_list = ObjectList(
    name="validation processes",
    get_objects=get_sorted_validations_list,
    filters=[
        DirectValuePropertyFilter(
            name="status",
            description="Filter list by current status of the process.",
            property_key="state",
            transform=transform_lower,
        ),
        DirectValuePropertyFilter(
            name="context-graph",
            description="Filter list by used data / context graph IRI.",
            property_key="contextGraphIri",
        ),
        DirectValuePropertyFilter(
            name="shape-graph",
            description="Filter list by used shape graph IRI.",
            property_key="shapeGraphIri",
        ),
        DirectValuePropertyFilter(
            name="more-resources-than",
            description="Filter list by the number of resources.",
            property_key="resourceCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
        DirectValuePropertyFilter(
            name="more-violations-than",
            description="Filter list by the number of violations.",
            property_key="violationsCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
        DirectValuePropertyFilter(
            name="more-violated-resources-than",
            description="Filter list by the number of violated resources.",
            property_key="resourcesWithViolationsCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
    ],
)


def get_violations_list(ctx: Context) -> list[dict]:
    """Get a list of violations"""
    try:
        # sometimes process_id is in params, sometimes in args !?
        process_id = ctx.params.get("process_id", None)
        if not process_id:
            process_id = ctx.args[0]
    except IndexError:
        return []  # process_id not given
    violations: list[dict] = []  # create a new object which better matches object_list needs
    for result in validation.get(batch_id=process_id)["results"]:
        resource_iri = result["resourceIri"]
        node_shapes = result["nodeShapes"]
        for _ in result["violations"]:
            _["resourceIri"] = resource_iri
            _["nodeShapes"] = node_shapes
            _["constraintName"] = _["reportEntryConstraintMessageTemplate"]["constraintName"]
            violations.append(_)
    return violations


violations_list = ObjectList(
    name="violations",
    get_objects=get_violations_list,
    filters=[
        DirectValuePropertyFilter(
            name="constraint",
            description="Filter list by constraint name.",
            property_key="constraintName",
        ),
        DirectValuePropertyFilter(
            name="severity", description="Filter list by severity.", property_key="severity"
        ),
        DirectValuePropertyFilter(
            name="resource", description="Filter list by resource IRI.", property_key="resourceIri"
        ),
        DirectListPropertyFilter(
            name="node-shape",
            description="Filter list by node shape IRI.",
            property_key="nodeShapes",
        ),
        DirectValuePropertyFilter(
            name="property-shape",
            description="Filter list by property shape IRI.",
            property_key="source",
        ),
    ],
)


def _get_batch_validation_option(validation_: dict) -> tuple[str, str]:
    """Get a completion option of a single batch validation"""
    id_ = validation_["id"]
    state = validation_["state"]
    graph = validation_["contextGraphIri"]
    stamp = datetime.fromtimestamp(validation_["executionStarted"] / 1000, tz=timezone.utc)
    time_ago = timeago.format(stamp, datetime.now(tz=timezone.utc))
    resources = _get_resource_count(validation_)
    violations = _get_violation_count(validation_)
    return (
        id_,
        f"{state} - {time_ago}, {resources} resources, {violations} violations ({graph})",
    )


def _complete_all_batch_validations(
    ctx: click.Context,  # noqa: ARG001
    param: click.Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Provide completion for batch validation"""
    options = [_get_batch_validation_option(_) for _ in validation.get_all_aggregations()]
    return _finalize_completion(candidates=options, incomplete=incomplete)


def _complete_running_batch_validations(
    ctx: click.Context,  # noqa: ARG001
    param: click.Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Provide completion for running batch validation"""
    options = [
        _get_batch_validation_option(_)
        for _ in validation.get_all_aggregations()
        if _["state"] == validation.STATUS_RUNNING
    ]
    return _finalize_completion(candidates=options, incomplete=incomplete)


def show_process_summary(app: ApplicationContext, process_id: str) -> None:
    """Show summary of the validation process"""
    app.echo_info_table(
        struct_to_table(validation.get_aggregation(batch_id=process_id)),
        headers=["Key", "Value"],
        sort_column=0,
        caption="Validation Summary",
    )


def show_violated_resources(app: ApplicationContext, data: list[dict]) -> None:
    """Show violated resource IRIs of a validation process"""
    messages = sorted({_["resourceIri"] for _ in data})
    app.echo_info(message=messages)


def _wait_for_process_completion(
    app: ApplicationContext, process_id: str, polling_interval: int, use_rich: bool = False
) -> str:
    class State:
        """State of a validation process"""

        id_: str
        data: dict
        status: str
        completed: int
        total: int

        def __init__(self, id_: str):
            self.id_ = id_
            self.refresh()

        def refresh(self) -> None:
            self.data = validation.get_aggregation(batch_id=self.id_)
            self.status = self.data.get("state", "UNKNOWN")
            self.completed = self.data.get("resourceProcessedCount", 0)
            self.total = self.data.get("resourceCount", 0)
            app.echo_debug(f"Process {self.id_} has status {self.status}.")

    state = State(id_=process_id)
    progress: Progress | None = None
    task: TaskID | None = None
    if use_rich:
        progress = Progress(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            transient=True,
            console=app.console,
        )
        progress.__enter__()  # simulate context manager (with:)
        task = progress.add_task(f"{state.status.capitalize()} ... ", total=state.total)
    while True:
        time.sleep(polling_interval)
        state.refresh()
        if progress is not None and task is not None:
            progress.update(
                task_id=task,
                completed=state.completed,
                description=f"{state.status.capitalize()} ... {state.completed} / {state.total}",
            )
        if state.status in (validation.STATUS_SCHEDULED, validation.STATUS_RUNNING):
            # when reported as running or scheduled, start another loop
            continue
        # when reported as finished, error or cancelled break out
        break
    if progress is not None and task is not None:
        progress.stop()
        progress.__exit__(None, None, None)
    if state.status == validation.STATUS_CANCELLED:
        raise ServerError("Process was cancelled.")
    if state.status == validation.STATUS_ERROR:
        error_message = state.data.get("error", "")
        raise ServerError(f"Process ended with error: {error_message}")
    return state.status


def _get_violation_table(violations: list[dict]) -> tuple[list, list]:
    """Get violation table from batch validation result"""
    table = []
    for violation in violations:
        resource_iri: str = str(violation.get("resourceIri"))
        path = violation.get("path", None)
        constraint_name = violation.get("constraintName", "UNKNOWN")
        node_shapes = violation.get("nodeShapes", [])
        text = violation["messages"][0]["value"]  # default: use the text of the first message
        for message in violation["messages"]:
            # look for en non non-lang messages to use
            if message["lang"] == "" or message["lang"] == "en":
                text = str(message["value"])
                break
        cell = ""
        if path is not None:
            cell = f"Path: {path}"
        if len(node_shapes) == 1:
            cell = f"{cell}\nNodeShape: {node_shapes[0]}"
        if len(node_shapes) > 1:
            cell = f"{cell}\nNodeShapes:"
            for node_shape in node_shapes:
                cell = f"{cell}\n - {node_shape}"
        cell = f"{cell}\nMessage: {text}"
        row = [resource_iri, constraint_name, cell]
        table.append(row)
    return table, ["Resource IRI", "Constraint", "Details"]


def _get_resource_count(batch_validation: dict) -> str:
    """Get resource count from validation report"""
    resource_count = str(batch_validation.get("resourceCount", "-"))
    processed_count = str(batch_validation.get("resourceProcessedCount", "-"))
    if resource_count == processed_count:
        return resource_count
    return f"{processed_count} / {resource_count}"


def _get_violation_count(process_data: dict) -> str:
    """Get violation count from validation report"""
    if process_data.get("executionStarted") is None:
        return "-"
    resources = str(process_data.get("resourcesWithViolationsCount", "0"))
    violations = str(process_data.get("violationsCount", "0"))
    if violations == "0":
        return "0"
    return f"{violations} ({resources} Resources)"


@click.command(cls=CmemcCommand, name="execute")
@click.argument("iri", type=click.STRING, shell_complete=completion.graph_uris)
@click.option(
    "--shape-graph",
    shell_complete=completion.graph_uris,
    default="https://vocab.eccenca.com/shacl/",
    show_default=True,
    help="The shape catalog used for validation.",
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Return the validation process identifier only. "
    "This is useful for piping the ID into other commands.",
)
@click.option(
    "--wait",
    is_flag=True,
    help="Wait until the process is finished. When using this option without the "
    "`--id-only` flag, it will enable a progress bar and a summary view.",
)
@click.option(
    "--polling-interval",
    type=click.IntRange(min=1),
    show_default=True,
    default=1,
    help="How many seconds to wait between status polls. Status polls are"
    " cheap, so a higher polling interval is most likely not needed.",
)
@click.pass_obj
def execute_command(  # noqa: PLR0913
    app: ApplicationContext,
    iri: str,
    shape_graph: str,
    id_only: bool,
    wait: bool,
    polling_interval: int,
) -> None:
    """Start a new validation process.

    Validation is performed on all typed resources of a data / context graph (IRI).
    Each resource is validated against all applicable node shapes from a
    selected shape catalog graph (and its sub-graphs).
    """
    process_id = validation.start(context_graph=iri, shape_graph=shape_graph)
    if wait:
        _wait_for_process_completion(
            app=app, process_id=process_id, use_rich=not id_only, polling_interval=polling_interval
        )
    if id_only:
        app.echo_info(process_id)
        return
    show_process_summary(process_id=process_id, app=app)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=validations_list.get_filter_help_text(),
    shell_complete=validations_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="List validation process identifier only. "
    "This is useful for piping the IDs into other commands.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the validation list.")
@click.pass_context
def list_command(ctx: Context, filter_: tuple[tuple[str, str]], id_only: bool, raw: bool) -> None:
    """List running and finished validation processes.

    This command provides a filterable table or identifier list of validation
    processes. The command operates on the process summary and provides some statistics.

    Note: Detailed information on the found violations can be listed with the
    `graph validation inspect` command.
    """
    validations = validations_list.apply_filters(ctx=ctx, filter_=filter_)
    app: ApplicationContext = ctx.obj

    if raw:
        app.echo_info_json(validations)
        return

    if id_only:
        for _ in validations:
            app.echo_info(_["id"])
        return

    if len(validations) == 0:
        app.echo_warning(
            "No validation processes found. "
            "Use `graph validation execute` to start a new validation process."
        )
        return

    # output a user table
    table = []
    for _ in validations:
        if "executionStarted" in _ and _["executionStarted"] is not None:
            stamp = datetime.fromtimestamp(_["executionStarted"] / 1000, tz=timezone.utc)
            time_ago = timeago.format(stamp, datetime.now(tz=timezone.utc))
        else:
            time_ago = f"{_['state']}"
        row = [
            _["id"],
            _["state"],
            time_ago,
            _["contextGraphIri"],
            _get_resource_count(_),
            _get_violation_count(_),
        ]
        table.append(row)
    app.echo_info_table(
        table, headers=["ID", "Status", "Started", "Graph", "Resources", "Violations"]
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("process_id", type=click.STRING, shell_complete=_complete_all_batch_validations)
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=violations_list.get_filter_help_text(),
    shell_complete=violations_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Return violated resource identifier only. "
    "This is useful for piping the ID into other commands.",
)
@click.option(
    "--summary",
    is_flag=True,
    help="Outputs the summary of the graph validation "
    "instead of the violations list (not filterable).",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the validation result.")
@click.pass_context
def inspect_command(  # noqa: PLR0913
    ctx: Context,
    process_id: str,
    filter_: tuple[tuple[str, str]],
    id_only: bool,
    summary: bool,
    raw: bool,
) -> None:
    """List and inspect errors found with a validation process.

    This command provides detailed information on the found violations of
    a validation process.

    Use the `--filter` option to limit the output based on different criteria such as
    constraint name (`constraint`), origin node shape of the rule (`node-shape`), or
    the validated resource (`resource`).

    Note: Validation processes IDs can be listed with the `graph validation list`
    command, or by utilizing the tab completion of this command.
    """
    app: ApplicationContext = ctx.obj
    if process_id not in [_["id"] for _ in validation.get_all_aggregations()]:
        raise UsageError(f"Validation process with ID '{process_id}' is not known (anymore).")

    if summary:
        if raw:
            app.echo_info_json(validation.get_aggregation(batch_id=process_id))
        else:
            show_process_summary(app=app, process_id=process_id)
        return

    data = violations_list.apply_filters(ctx=ctx, filter_=filter_)

    if id_only:
        show_violated_resources(app=app, data=data)
        return

    if raw:
        app.echo_info_json(data)
        return

    if len(data) == 0 and len(filter_) == 0:
        app.echo_warning(
            "The given validation process does not have any violations - "
            "I will show the summary instead."
        )
        show_process_summary(app=app, process_id=process_id)
    else:
        messages_table, messages_header = _get_violation_table(violations=data)
        if len(messages_table) > 0:
            app.echo_info("")
            app.echo_info_table(messages_table, headers=messages_header, sort_column=0)


@click.command(cls=CmemcCommand, name="cancel")
@click.argument("process_id", type=click.STRING, shell_complete=_complete_running_batch_validations)
@click.pass_obj
def cancel_command(app: ApplicationContext, process_id: str) -> None:
    """Cancel a running validation process.

    Note: In order to get the process IDs of all currently running validation
    processes, use the `graph validation list` command with the option
    `--filter status running`, or utilize the tab completion of this command.
    """
    try:
        graph_validation = validation.get_aggregation(process_id)
    except HTTPError as error:
        if error.response is not None and error.response.status_code == requests.codes.not_found:
            raise click.UsageError(
                f"Graph validation process with ID {process_id} does not exist."
            ) from error
    else:
        if graph_validation["state"] != validation.STATUS_RUNNING:
            raise click.UsageError(
                f"Graph validation process with ID {process_id} is not a running anymore."
            )
    app.echo_info(f"Graph validation process with ID {process_id} ... ", nl=False)
    validation.cancel(batch_id=process_id)
    app.echo_success("cancelled")


@click.group(cls=CmemcGroup, name="validation")
def validation_group() -> CmemcGroup:  # type: ignore[empty-body]
    """Validate resources in a graph.

    This command group is dedicated to the management of resource validation processes.
    A validation process verifies, that resources in a specific graph are valid according
    to the node shapes in a shape catalog graph.

    Note: Validation processes are identified with a random ID and can be listed with
    the `graph validation list` command. To start or cancel validation processes,
    use the `graph validation execute` and `graph validation cancel` command.
    To inspect the found violations of a validation process, use the
    `graph validation inspect` command.
    """


validation_group.add_command(execute_command)
validation_group.add_command(list_command)
validation_group.add_command(inspect_command)
validation_group.add_command(cancel_command)
